##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Verifier that checks a propfind response for regex matches to property values.
"""

from xml.etree.cElementTree import ElementTree, tostring
from StringIO import StringIO
import re
import urllib

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable

        # If no status verification requested, then assume all 2xx codes are OK
        ignores = args.get("ignore", [])
        only = args.get("only", [])

        def normalizeXML(value):

            if value[0] == '<':
                try:
                    tree = ElementTree(file=StringIO(value))
                except Exception:
                    return False, "           Could not parse XML value: %s\n" % (value,)
                value = tostring(tree.getroot())
            return value

        # Get property arguments and split on $ delimited for name, value tuples
        testprops = args.get("props", [])
        props_match = []
        for i in range(len(testprops)):
            p = testprops[i]
            if (p.find("$") != -1):
                if p.find("$") != len(p) - 1:
                    props_match.append((p.split("$")[0], normalizeXML(p.split("$")[1]), True))
                else:
                    props_match.append((p.split("$")[0], "", True))
            elif (p.find("!") != -1):
                if p.find("!") != len(p) - 1:
                    props_match.append((p.split("!")[0], normalizeXML(p.split("!")[1]), False))
                else:
                    props_match.append((p.split("!")[0], "", False))

        # Process the multistatus response, extracting all hrefs
        # and comparing with the properties defined for this test. Report any
        # mismatches.

        # Must have MULTISTATUS response code
        if response.status != 207:
            return False, "           HTTP Status for Request: %d\n" % (response.status,)

        try:
            tree = ElementTree(file=StringIO(respdata))
        except Exception:
            return False, "           Could not parse proper XML response\n"

        result = True
        resulttxt = ""
        for response in tree.findall("{DAV:}response"):

            # Get href for this response
            href = response.findall("{DAV:}href")
            if len(href) != 1:
                return False, "           Wrong number of DAV:href elements\n"
            href = urllib.unquote(href[0].text)
            if href in ignores:
                continue
            if only and href not in only:
                continue

            # Get all property status
            ok_status_props = {}
            propstatus = response.findall("{DAV:}propstat")
            for props in propstatus:
                # Determine status for this propstat
                status = props.findall("{DAV:}status")
                if len(status) == 1:
                    statustxt = status[0].text
                    status = False
                    if statustxt.startswith("HTTP/1.1 ") and (len(statustxt) >= 10):
                        status = (statustxt[9] == "2")
                else:
                    status = False

                # Get properties for this propstat
                prop = props.findall("{DAV:}prop")
                if len(prop) != 1:
                    return False, "           Wrong number of DAV:prop elements\n"

                def _removeWhitespace(node):

                    for child in node.getchildren():
                        child.text = child.text.strip() if child.text else child.text
                        child.tail = child.tail.strip() if child.tail else child.tail
                        _removeWhitespace(child)

                for child in prop[0].getchildren():
                    fqname = child.tag
                    if len(child):
                        value = ""
                        _removeWhitespace(child)
                        for p in child.getchildren():
                            value += tostring(p)
                    elif child.text:
                        value = child.text
                    else:
                        value = None

                    if status:
                        ok_status_props[fqname] = value

            # Look at each property we want to test and see if present
            for propname, value, match in props_match:
                if propname not in ok_status_props:
                    resulttxt += "        Items not returned in report (OK) for %s: %s\n" % (href, propname,)
                    result = False
                    continue
                matched = re.match(value, ok_status_props[propname])
                if match and not matched:
                    resulttxt += "        Items not matching for %s: %s %s\n" % (href, propname, ok_status_props[propname])
                    result = False
                elif not match and matched:
                    resulttxt += "        Items incorrectly match for %s: %s %s\n" % (href, propname, ok_status_props[propname])
                    result = False

        return result, resulttxt



##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from verifiers.addressDataMatch import Verifier as CardVerifier

"""
Verifier that checks the response body for a semantic match to data in a file.
"""

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args):

        # Just hand this off to the addressDataMatch verifier which knows all about
        # proper vCard normalization rules.
        verifier = CardVerifier()
        return verifier.verify(manager, uri, response, respdata, args, is_json=True)



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##
from uuid import uuid4

"""
Class that encapsulates the server information for a CalDAV test run.
"""

import datetime
import re
import src.xmlDefs


class serverinfo(object):
    """
    Maintains information about the server being targeted.
    """

    # RegEx pattern to match substitution variables
    subspattern = re.compile("(?P<name>\\$[_a-zA-Z][_a-zA-Z0-9\\-]*\\:)")

    def __init__(self):
        self.host = ""
        self.nonsslport = 80
        self.sslport = 443
        self.afunix = None
        self.host2 = ""
        self.nonsslport2 = 80
        self.sslport2 = 443
        self.afunix2 = None
        self.authtype = "basic"
        self.features = set()
        self.user = ""
        self.pswd = ""
        self.waitcount = 120
        self.waitdelay = 0.25
        self.waitsuccess = 10
        self.subsdict = {}
        self.extrasubsdict = {}

        # dtnow needs to be fixed to a single date at the start of the tests just in case the tests
        # run over a day boundary.
        self.dtnow = datetime.date.today()


    def _re_subs(self, sub, mapping):
        """
        Do a regex substitution via the supplied mapping, only if the mapping exists.

        @param sub: string to do substitution in
        @type sub: L{str}
        @param mapping: mapping of substitution name to value
        @type mapping: L{dict}
        """
        # Helper function for .sub()
        def convert(mo):
            named = mo.group('name')
            if named is not None and named in mapping:
                return mapping[named]
            else:
                return named
        return self.subspattern.sub(convert, sub)


    def subs(self, sub, db=None):

        # Special handling for relative date-times
        pos = sub.find("$now.")
        while pos != -1:
            endpos = pos + sub[pos:].find(":")
            if sub[pos:].startswith("$now.year."):
                yearoffset = int(sub[pos + 10:endpos])
                value = "%d" % (self.dtnow.year + yearoffset,)
            elif sub[pos:].startswith("$now.month."):
                monthoffset = int(sub[pos + 11:endpos])
                month = self.dtnow.month + monthoffset
                year = self.dtnow.year + divmod(month - 1, 12)[0]
                month = divmod(month - 1, 12)[1] + 1
                value = "%d%02d" % (year, month,)
            elif sub[pos:].startswith("$now.week."):
                weekoffset = int(sub[pos + 10:endpos])
                dtoffset = self.dtnow + datetime.timedelta(days=7 * weekoffset)
                value = "%d%02d%02d" % (dtoffset.year, dtoffset.month, dtoffset.day,)
            else:
                dayoffset = int(sub[pos + 5:endpos])
                dtoffset = self.dtnow + datetime.timedelta(days=dayoffset)
                value = "%d%02d%02d" % (dtoffset.year, dtoffset.month, dtoffset.day,)
            sub = "%s%s%s" % (sub[:pos], value, sub[endpos + 1:])
            pos = sub.find("$now.")

        if sub.find("$uidrandom:") != -1:
            sub = sub.replace("$uidrandom:", str(uuid4()))

        if db is None:
            db = self.subsdict
        while '$' in sub:
            newstr = self._re_subs(sub, db)
            if newstr == sub:
                break
            sub = newstr
        return sub


    def addsubs(self, items, db=None):
        if db is None:
            db_actual = self.subsdict
        else:
            db_actual = db
        for key, value in items.iteritems():
            db_actual[key] = value

        if db is None:
            self.updateParams()


    def hasextrasubs(self):
        return len(self.extrasubsdict) > 0


    def extrasubs(self, str):
        return self.subs(str, self.extrasubsdict)


    def addextrasubs(self, items):
        processed = {}

        # Various "functions" might be applied to a variable name to cause the value to
        # be changed in various ways
        for variable, value in items.items():

            # basename() - extract just the URL last path segment from the value
            if variable.startswith("basename("):
                variable = variable[len("basename("):-1]
                value = value.rstrip("/").split("/")[-1]
            processed[variable] = value

        self.addsubs(processed, self.extrasubsdict)


    def newUIDs(self):
        uidsubs = dict([("$uid{u}:".format(u=i), str(uuid4())) for i in range(1, 21)])
        self.subsdict.update(uidsubs)
        self.extrasubsdict.update(uidsubs)
        return set([(v, k) for k, v in uidsubs.items()])


    def parseXML(self, node):
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_HOST:
                try:
                    self.host = child.text.encode("utf-8")
                except:
                    self.host = "localhost"
            elif child.tag == src.xmlDefs.ELEMENT_NONSSLPORT:
                self.nonsslport = int(child.text)
            elif child.tag == src.xmlDefs.ELEMENT_SSLPORT:
                self.sslport = int(child.text)
            elif child.tag == src.xmlDefs.ELEMENT_UNIX:
                self.afunix = child.text
            elif child.tag == src.xmlDefs.ELEMENT_HOST2:
                try:
                    self.host2 = child.text.encode("utf-8")
                except:
                    self.host2 = "localhost"
            elif child.tag == src.xmlDefs.ELEMENT_NONSSLPORT2:
                self.nonsslport2 = int(child.text)
            elif child.tag == src.xmlDefs.ELEMENT_SSLPORT2:
                self.sslport2 = int(child.text)
            elif child.tag == src.xmlDefs.ELEMENT_UNIX2:
                self.afunix2 = child.text
            elif child.tag == src.xmlDefs.ELEMENT_AUTHTYPE:
                self.authtype = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_WAITCOUNT:
                self.waitcount = int(child.text.encode("utf-8"))
            elif child.tag == src.xmlDefs.ELEMENT_WAITDELAY:
                self.waitdelay = float(child.text.encode("utf-8"))
            elif child.tag == src.xmlDefs.ELEMENT_WAITSUCCESS:
                self.waitsuccess = int(child.text.encode("utf-8"))
            elif child.tag == src.xmlDefs.ELEMENT_FEATURES:
                self.parseFeatures(child)
            elif child.tag == src.xmlDefs.ELEMENT_SUBSTITUTIONS:
                self.parseSubstitutionsXML(child)

        self.updateParams()


    def parseFeatures(self, node):
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_FEATURE:
                self.features.add(child.text.encode("utf-8"))


    def updateParams(self):

        # Expand substitutions fully at this point
        for k, v in self.subsdict.items():
            while '$' in v:
                v = self._re_subs(v, self.subsdict)
            self.subsdict[k] = v

        # Now cache some useful substitutions
        user = "$userid1:" if "$userid1:" in self.subsdict else "$userid01:"
        pswd = "$pswd1:" if "$pswd1:" in self.subsdict else "$pswd01:"
        if user not in self.subsdict:
            raise ValueError("Must have userid substitution")
        self.user = self.subsdict[user]
        if pswd not in self.subsdict:
            raise ValueError("Must have pswd substitution")
        self.pswd = self.subsdict[pswd]


    def parseRepeatXML(self, node):
        # Look for count
        count = node.get(src.xmlDefs.ATTR_COUNT)

        for child in node.getchildren():
            self.parseSubstitutionXML(child, count)


    def parseSubstitutionsXML(self, node):
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_SUBSTITUTION:
                self.parseSubstitutionXML(child)
            elif child.tag == src.xmlDefs.ELEMENT_REPEAT:
                self.parseRepeatXML(child)


    def parseSubstitutionXML(self, node, repeat=None):
        if node.tag == src.xmlDefs.ELEMENT_SUBSTITUTION:
            key = None
            value = None
            for schild in node.getchildren():
                if schild.tag == src.xmlDefs.ELEMENT_KEY:
                    key = schild.text.encode("utf-8")
                elif schild.tag == src.xmlDefs.ELEMENT_VALUE:
                    value = schild.text.encode("utf-8") if schild.text else ""

            if key and value:
                if repeat:
                    for count in range(1, int(repeat) + 1):
                        self.subsdict[key % (count,)] = (value % (count,)) if "%" in value else value
                else:
                    self.subsdict[key] = value



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Verifier that chec ks the response status code for a specific value.
"""

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable
        # If no status verification requested, then assume all 2xx codes are OK
        teststatus = args.get("status", ["2xx"])

        for test in teststatus:
            if test[1:3] == "xx":
                test = int(test[0])
            else:
                test = int(test)
            if test < 100:
                result = ((response.status / 100) == test)
            else:
                result = (response.status == test)
            if result:
                return True, ""

        return False, "        HTTP Status Code Wrong: %d" % (response.status,)



##
# Copyright (c) 2010-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

import httplib
import socket
import ssl as sslmodule

class HTTPSVersionConnection(httplib.HTTPSConnection):
    """
    An L{httplib.HTTPSConnection} class that allows the TLS protocol version to be set.
    """

    def __init__(self, host, port, ssl_version=sslmodule.PROTOCOL_TLSv1):
        httplib.HTTPSConnection.__init__(self, host, port)
        self._ssl_version = ssl_version


    def connect(self):
        "Connect to a host on a given (SSL) port."

        sock = socket.create_connection((self.host, self.port), self.timeout)
        self.sock = sslmodule.wrap_socket(sock, self.key_file, self.cert_file, ssl_version=self._ssl_version)


cached_types = (
    (set(), sslmodule.PROTOCOL_TLSv1),
    (set(), sslmodule.PROTOCOL_SSLv3),
    (set(), sslmodule.PROTOCOL_SSLv23),
)

class UnixSocketHTTPConnection(httplib.HTTPConnection):
    """
    An L{httplib.HTTPConnection} class that uses a unix socket rather than TCP.
    """

    def __init__(self, path, host, port):
        httplib.HTTPConnection.__init__(self, host, port)
        self.path = path


    def connect(self):
        """
        Connect using the supplied unix socket file path
        """
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(self.path)



def SmartHTTPConnection(host, port, ssl, afunix):
    """
    Create the appropriate L{httplib.HTTPConnection} derived class for the supplied arguments.

    @param host: TCP host name
    @type host: L{str}
    @param port: TCP port number
    @type port: L{int}
    @param ssl: indicates if SSL is to be used
    @type ssl: L{bool}
    @param afunix: unix socket to use or L{None}
    @type afunix: L{str}
    """

    def trySSL(version):
        connect = HTTPSVersionConnection(host, port, ssl_version=version)
        connect.connect()
        return connect

    if afunix:
        connect = UnixSocketHTTPConnection(afunix, host, port)
    elif ssl:
        # Iterate over the TL:S versions and find one that works and cache it for future use.
        for cached, connection_type in cached_types:
            if (host, port) in cached:
                try:
                    return trySSL(connection_type)
                except:
                    cached.remove((host, port))

        for cached, connection_type in cached_types:
            try:
                cached.add((host, port))
                return trySSL(connection_type)
            except:
                cached.remove((host, port))

        raise RuntimeError("Cannot connect via with TLSv1, SSLv3 or SSLv23")
    else:
        connect = httplib.HTTPConnection(host, port)
    connect.connect()
    return connect



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Class that encapsulates a series of tests.
"""

from src.test import test
from src.xmlUtils import getYesNoAttributeValue
import src.xmlDefs

class testsuite(object):
    """
    Maintains a list of tests to run as part of a 'suite'.
    """

    def __init__(self, manager):
        self.manager = manager
        self.name = ""
        self.ignore = False
        self.only = False
        self.changeuid = False
        self.require_features = set()
        self.exclude_features = set()
        self.tests = []


    def aboutToRun(self):
        """
        Typically we need the calendar/contact data for a test file to have a common set
        of UIDs, and for each overall test file to have unique UIDs. Occasionally, within
        a test file we also need test suites to have unique UIDs. The "change-uid" attribute
        can be used to reset the active UIDs for a test suite.
        """
        return self.manager.server_info.newUIDs() if self.changeuid else set()


    def missingFeatures(self):
        return self.require_features - self.manager.server_info.features


    def excludedFeatures(self):
        return self.exclude_features & self.manager.server_info.features


    def parseXML(self, node):
        self.name = node.get(src.xmlDefs.ATTR_NAME, "")
        self.ignore = getYesNoAttributeValue(node, src.xmlDefs.ATTR_IGNORE)
        self.only = getYesNoAttributeValue(node, src.xmlDefs.ATTR_ONLY)
        self.changeuid = getYesNoAttributeValue(node, src.xmlDefs.ATTR_CHANGE_UID)

        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_REQUIRE_FEATURE:
                self.parseFeatures(child, require=True)
            elif child.tag == src.xmlDefs.ELEMENT_EXCLUDE_FEATURE:
                self.parseFeatures(child, require=False)
            elif child.tag == src.xmlDefs.ELEMENT_TEST:
                t = test(self.manager)
                t.parseXML(child)
                self.tests.append(t)


    def parseFeatures(self, node, require=True):
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_FEATURE:
                (self.require_features if require else self.exclude_features).add(child.text.encode("utf-8"))


    def dump(self):
        print "\nTest Suite:"
        print "    name: %s" % self.name
        for iter in self.tests:
            iter.dump()



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Defines the 'request' class which encapsulates an HTTP request and verification.
"""

from hashlib import md5, sha1
from src.httpshandler import SmartHTTPConnection
from src.xmlUtils import getYesNoAttributeValue
import base64
import datetime
import os
import re
import src.xmlDefs
import time
import urllib
import urlparse
import uuid

algorithms = {
    'md5': md5,
    'md5-sess': md5,
    'sha': sha1,
}

# DigestCalcHA1
def calcHA1(
    pszAlg,
    pszUserName,
    pszRealm,
    pszPassword,
    pszNonce,
    pszCNonce,
    preHA1=None
):
    """
    @param pszAlg: The name of the algorithm to use to calculate the digest.
        Currently supported are md5 md5-sess and sha.

    @param pszUserName: The username
    @param pszRealm: The realm
    @param pszPassword: The password
    @param pszNonce: The nonce
    @param pszCNonce: The cnonce

    @param preHA1: If available this is a str containing a previously
       calculated HA1 as a hex string. If this is given then the values for
       pszUserName, pszRealm, and pszPassword are ignored.
    """

    if (preHA1 and (pszUserName or pszRealm or pszPassword)):
        raise TypeError(("preHA1 is incompatible with the pszUserName, "
                         "pszRealm, and pszPassword arguments"))

    if preHA1 is None:
        # We need to calculate the HA1 from the username:realm:password
        m = algorithms[pszAlg]()
        m.update(pszUserName)
        m.update(":")
        m.update(pszRealm)
        m.update(":")
        m.update(pszPassword)
        HA1 = m.digest()
    else:
        # We were given a username:realm:password
        HA1 = preHA1.decode('hex')

    if pszAlg == "md5-sess":
        m = algorithms[pszAlg]()
        m.update(HA1)
        m.update(":")
        m.update(pszNonce)
        m.update(":")
        m.update(pszCNonce)
        HA1 = m.digest()

    return HA1.encode('hex')



# DigestCalcResponse
def calcResponse(
    HA1,
    algo,
    pszNonce,
    pszNonceCount,
    pszCNonce,
    pszQop,
    pszMethod,
    pszDigestUri,
    pszHEntity,
):
    m = algorithms[algo]()
    m.update(pszMethod)
    m.update(":")
    m.update(pszDigestUri)
    if pszQop == "auth-int":
        m.update(":")
        m.update(pszHEntity)
    HA2 = m.digest().encode('hex')

    m = algorithms[algo]()
    m.update(HA1)
    m.update(":")
    m.update(pszNonce)
    m.update(":")
    if pszNonceCount and pszCNonce and pszQop:
        m.update(pszNonceCount)
        m.update(":")
        m.update(pszCNonce)
        m.update(":")
        m.update(pszQop)
        m.update(":")
    m.update(HA2)
    respHash = m.digest().encode('hex')
    return respHash



class pause (object):
    pass



class request(object):
    """
    Represents the HTTP request to be executed, and verification information to
    be used to determine a satisfactory output or not.
    """

    nc = {} # Keep track of nonce count

    def __init__(self, manager):
        self.manager = manager
        self.host = self.manager.server_info.host
        self.port = self.manager.server_info.port
        self.afunix = self.manager.server_info.afunix
        self.auth = True
        self.user = ""
        self.pswd = ""
        self.end_delete = False
        self.print_request = False
        self.print_response = False
        self.wait_for_success = None
        self.require_features = set()
        self.exclude_features = set()
        self.method = ""
        self.headers = {}
        self.ruris = []
        self.ruri = ""
        self.ruri_quote = True
        self.data = None
        self.iterate_data = False
        self.count = 1
        self.verifiers = []
        self.graburi = None
        self.grabcount = None
        self.grabheader = []
        self.grabproperty = []
        self.grabelement = []
        self.grabjson = []
        self.grabcalprop = []
        self.grabcalparam = []


    def __str__(self):
        return "Method: %s; uris: %s" % (self.method, self.ruris if len(self.ruris) > 1 else self.ruri,)


    def missingFeatures(self):
        return self.require_features - self.manager.server_info.features


    def excludedFeatures(self):
        return self.exclude_features & self.manager.server_info.features


    def getURI(self, si):
        uri = si.extrasubs(self.ruri)
        if "**" in uri:
            if "?" not in uri or uri.find("?") > uri.find("**"):
                uri = uri.replace("**", str(uuid.uuid4()))
        elif "##" in uri:
            if "?" not in uri or uri.find("?") > uri.find("##"):
                uri = uri.replace("##", str(self.count))
        return uri


    def getHeaders(self, si):
        hdrs = self.headers
        for key, value in hdrs.items():
            hdrs[key] = si.extrasubs(value)

        # Content type
        if self.data != None:
            hdrs["Content-Type"] = self.data.content_type

        # Auth
        if self.auth:
            if si.authtype.lower() == "digest":
                hdrs["Authorization"] = self.gethttpdigestauth(si)
            else:
                hdrs["Authorization"] = self.gethttpbasicauth(si)

        return hdrs


    def gethttpbasicauth(self, si):
        basicauth = [self.user, si.user][self.user == ""]
        basicauth += ":"
        basicauth += [self.pswd, si.pswd][self.pswd == ""]
        basicauth = "Basic " + base64.encodestring(basicauth)
        basicauth = basicauth.replace("\n", "")
        return basicauth


    def gethttpdigestauth(self, si, wwwauthorize=None):

        # Check the nonce cache to see if we've used this user before, or if the nonce is more than 5 minutes old
        user = [self.user, si.user][self.user == ""]
        pswd = [self.pswd, si.pswd][self.pswd == ""]
        details = None
        if user in self.manager.digestCache and self.manager.digestCache[user]["max-nonce-time"] > time.time():
            details = self.manager.digestCache[user]
        else:
            # Redo digest auth from scratch to get a new nonce etc
            http = SmartHTTPConnection(si.host, si.port, si.ssl, si.afunix)
            try:
                puri = list(urlparse.urlparse(self.getURI(si)))
                puri[2] = urllib.quote(puri[2])
                quri = urlparse.urlunparse(puri)
                http.request("OPTIONS", quri)

                response = http.getresponse()

            finally:
                http.close()

            if response.status == 401:

                wwwauthorize = response.msg.getheaders("WWW-Authenticate")
                for item in wwwauthorize:
                    if not item.lower().startswith("digest "):
                        continue
                    wwwauthorize = item[7:]
                    def unq(s):
                        if s[0] == s[-1] == '"':
                            return s[1:-1]
                        return s
                    parts = wwwauthorize.split(',')

                    details = {}

                    for (k, v) in [p.split('=', 1) for p in parts]:
                        details[k.strip()] = unq(v.strip())

                    details["max-nonce-time"] = time.time() + 600
                    self.manager.digestCache[user] = details
                    break

        if details:
            if details.get('qop'):
                if self.nc.get(details.get('nonce')) is None:
                    self.nc[details.get('nonce')] = 1
                else:
                    self.nc[details.get('nonce')] += 1
                details['nc'] = "%08x" % self.nc[details.get('nonce')]
                if details.get('cnonce') is None:
                    details['cnonce'] = "D4AAE4FF-ADA1-4149-BFE2-B506F9264318"

            digest = calcResponse(
                calcHA1(details.get('algorithm', 'md5'), user, details.get('realm'), pswd, details.get('nonce'), details.get('cnonce')),
                details.get('algorithm', 'md5'), details.get('nonce'), details.get('nc'), details.get('cnonce'), details.get('qop'), self.method, self.getURI(si), None
            )

            if details.get('qop'):
                response = (
                    'Digest username="%s", realm="%s", '
                    'nonce="%s", uri="%s", '
                    'response=%s, algorithm=%s, cnonce="%s", qop=%s, nc=%s' %
                    (user, details.get('realm'), details.get('nonce'), self.getURI(si), digest, details.get('algorithm', 'md5'), details.get('cnonce'), details.get('qop'), details.get('nc'),)
                )
            else:
                response = (
                    'Digest username="%s", realm="%s", '
                    'nonce="%s", uri="%s", '
                    'response=%s, algorithm=%s' %
                    (user, details.get('realm'), details.get('nonce'), self.getURI(si), digest, details.get('algorithm'),)
                )

            return response
        else:
            return ""


    def getFilePath(self):
        if self.data != None:
            return os.path.join(self.manager.data_dir, self.data.filepath) if self.manager.data_dir else self.data.filepath
        else:
            return ""


    def getData(self):
        data = ""
        if self.data != None:
            if len(self.data.value) != 0:
                data = self.data.value
            elif self.data.filepath:
                # read in the file data
                fd = open(self.data.nextpath if hasattr(self.data, "nextpath") else self.getFilePath(), "r")
                try:
                    data = fd.read()
                finally:
                    fd.close()
            data = str(self.manager.server_info.subs(data))
            self.manager.server_info.addextrasubs({"$request_count:": str(self.count)})
            data = self.manager.server_info.extrasubs(data)
            if self.data.substitutions:
                data = self.manager.server_info.subs(data, self.data.substitutions)
            if self.data.generate:
                if self.data.content_type.startswith("text/calendar"):
                    data = self.generateCalendarData(data)
            elif self.data.generator:
                data = self.data.generator.doGenerate()
        return data


    def getNextData(self):
        if not hasattr(self, "dataList"):
            self.dataList = sorted([path for path in os.listdir(self.getFilePath()) if not path.startswith(".")])
        if len(self.dataList):
            self.data.nextpath = os.path.join(self.getFilePath(), self.dataList.pop(0))
            return True
        else:
            if hasattr(self.data, "nextpath"):
                delattr(self.data, "nextpath")
            if hasattr(self, "dataList"):
                delattr(self, "dataList")
            return False


    def hasNextData(self):
        dataList = sorted([path for path in os.listdir(self.getFilePath()) if not path.startswith(".")])
        return len(dataList) != 0


    def generateCalendarData(self, data):
        """
        FIXME: does not work for events with recurrence overrides.
        """

        # Change the following iCalendar data values:
        # DTSTART, DTEND, RECURRENCE-ID, UID

        data = re.sub("UID:.*", "UID:%s" % (uuid.uuid4(),), data)
        data = re.sub("SUMMARY:(.*)", "SUMMARY:\\1 #%s" % (self.count,), data)

        now = datetime.date.today()
        data = re.sub("(DTSTART;[^:]*):[0-9]{8,8}", "\\1:%04d%02d%02d" % (now.year, now.month, now.day,), data)
        data = re.sub("(DTEND;[^:]*):[0-9]{8,8}", "\\1:%04d%02d%02d" % (now.year, now.month, now.day,), data)

        return data


    def parseXML(self, node):
        self.auth = node.get(src.xmlDefs.ATTR_AUTH, src.xmlDefs.ATTR_VALUE_YES) == src.xmlDefs.ATTR_VALUE_YES
        self.user = self.manager.server_info.subs(node.get(src.xmlDefs.ATTR_USER, "").encode("utf-8"))
        self.pswd = self.manager.server_info.subs(node.get(src.xmlDefs.ATTR_PSWD, "").encode("utf-8"))
        self.end_delete = getYesNoAttributeValue(node, src.xmlDefs.ATTR_END_DELETE)
        self.print_request = self.manager.print_request or getYesNoAttributeValue(node, src.xmlDefs.ATTR_PRINT_REQUEST)
        self.print_response = self.manager.print_response or getYesNoAttributeValue(node, src.xmlDefs.ATTR_PRINT_RESPONSE)
        self.iterate_data = getYesNoAttributeValue(node, src.xmlDefs.ATTR_ITERATE_DATA)
        self.wait_for_success = getYesNoAttributeValue(node, src.xmlDefs.ATTR_WAIT_FOR_SUCCESS)

        if node.get(src.xmlDefs.ATTR_HOST2, src.xmlDefs.ATTR_VALUE_NO) == src.xmlDefs.ATTR_VALUE_YES:
            self.host = self.manager.server_info.host2
            self.port = self.manager.server_info.port2
            self.afunix = self.manager.server_info.afunix2

        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_REQUIRE_FEATURE:
                self.parseFeatures(child, require=True)
            elif child.tag == src.xmlDefs.ELEMENT_EXCLUDE_FEATURE:
                self.parseFeatures(child, require=False)
            elif child.tag == src.xmlDefs.ELEMENT_METHOD:
                self.method = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_HEADER:
                self.parseHeader(child)
            elif child.tag == src.xmlDefs.ELEMENT_RURI:
                self.ruri_quote = child.get(src.xmlDefs.ATTR_QUOTE, src.xmlDefs.ATTR_VALUE_YES) == src.xmlDefs.ATTR_VALUE_YES
                self.ruris.append(self.manager.server_info.subs(child.text.encode("utf-8")))
                if len(self.ruris) == 1:
                    self.ruri = self.ruris[0]
            elif child.tag == src.xmlDefs.ELEMENT_DATA:
                self.data = data(self.manager)
                self.data.parseXML(child)
            elif child.tag == src.xmlDefs.ELEMENT_VERIFY:
                self.verifiers.append(verify(self.manager))
                self.verifiers[-1].parseXML(child)
            elif child.tag == src.xmlDefs.ELEMENT_GRABURI:
                self.graburi = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_GRABCOUNT:
                self.grabcount = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_GRABHEADER:
                self.parseGrab(child, self.grabheader)
            elif child.tag == src.xmlDefs.ELEMENT_GRABPROPERTY:
                self.parseGrab(child, self.grabproperty)
            elif child.tag == src.xmlDefs.ELEMENT_GRABELEMENT:
                self.parseMultiGrab(child, self.grabelement)
            elif child.tag == src.xmlDefs.ELEMENT_GRABJSON:
                self.parseMultiGrab(child, self.grabjson)
            elif child.tag == src.xmlDefs.ELEMENT_GRABCALPROP:
                self.parseGrab(child, self.grabcalprop)
            elif child.tag == src.xmlDefs.ELEMENT_GRABCALPARAM:
                self.parseGrab(child, self.grabcalparam)


    def parseFeatures(self, node, require=True):
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_FEATURE:
                (self.require_features if require else self.exclude_features).add(child.text.encode("utf-8"))


    def parseHeader(self, node):

        name = None
        value = None
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_NAME:
                name = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_VALUE:
                value = self.manager.server_info.subs(child.text.encode("utf-8"))

        if (name is not None) and (value is not None):
            self.headers[name] = value


    def parseList(manager, node):
        requests = []
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_REQUEST:
                req = request(manager)
                req.parseXML(child)
                requests.append(req)
            elif child.tag == src.xmlDefs.ELEMENT_PAUSE:
                requests.append(pause())
        return requests

    parseList = staticmethod(parseList)

    def parseGrab(self, node, appendto):

        name = None
        variable = None
        for child in node.getchildren():
            if child.tag in (src.xmlDefs.ELEMENT_NAME, src.xmlDefs.ELEMENT_PROPERTY):
                name = self.manager.server_info.subs(child.text.encode("utf-8"))
            elif child.tag == src.xmlDefs.ELEMENT_VARIABLE:
                variable = self.manager.server_info.subs(child.text.encode("utf-8"))

        if (name is not None) and (variable is not None):
            appendto.append((name, variable))


    def parseMultiGrab(self, node, appendto):

        name = None
        parent = None
        variable = None
        for child in node.getchildren():
            if child.tag in (src.xmlDefs.ELEMENT_NAME, src.xmlDefs.ELEMENT_PROPERTY, src.xmlDefs.ELEMENT_POINTER):
                name = self.manager.server_info.subs(child.text.encode("utf-8"))
            elif child.tag == src.xmlDefs.ELEMENT_PARENT:
                parent = self.manager.server_info.subs(child.text.encode("utf-8"))
            elif child.tag == src.xmlDefs.ELEMENT_VARIABLE:
                if variable is None:
                    variable = []
                variable.append(self.manager.server_info.subs(child.text.encode("utf-8")))

        if (name is not None) and (variable is not None):
            appendto.append((name, variable,) if parent is None else (name, parent, variable,))



class data(object):
    """
    Represents the data/body portion of an HTTP request.
    """

    def __init__(self, manager):
        self.manager = manager
        self.content_type = ""
        self.filepath = ""
        self.generator = None
        self.value = ""
        self.substitutions = {}
        self.substitute = False
        self.generate = False


    def parseXML(self, node):

        self.substitute = node.get(src.xmlDefs.ATTR_SUBSTITUTIONS, src.xmlDefs.ATTR_VALUE_YES) == src.xmlDefs.ATTR_VALUE_YES
        self.generate = node.get(src.xmlDefs.ATTR_GENERATE, src.xmlDefs.ATTR_VALUE_NO) == src.xmlDefs.ATTR_VALUE_YES

        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_CONTENTTYPE:
                self.content_type = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_FILEPATH:
                self.filepath = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_GENERATOR:
                self.generator = generator(self.manager)
                self.generator.parseXML(child)
            elif child.tag == src.xmlDefs.ELEMENT_SUBSTITUTE:
                self.parseSubstituteXML(child)


    def parseSubstituteXML(self, node):
        name = None
        value = None
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_NAME:
                name = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_VALUE:
                value = self.manager.server_info.subs(child.text.encode("utf-8"))
        if name and value:
            self.substitutions[name] = value



class generator(object):
    """
    Defines a dynamically generated request body.
    """

    def __init__(self, manager):
        self.manager = manager
        self.callback = None
        self.args = {}


    def doGenerate(self):

        # Re-do substitutions from values generated during the current test run
        if self.manager.server_info.hasextrasubs():
            for name, values in self.args.iteritems():
                newvalues = [self.manager.server_info.extrasubs(value) for value in values]
                self.args[name] = newvalues

        generatorClass = self._importName(self.callback, "Generator")
        gen = generatorClass()

        # Always clone the args as this verifier may be called multiple times
        args = dict((k, list(v)) for k, v in self.args.items())

        return gen.generate(self.manager, args)


    def _importName(self, modulename, name):
        """
        Import a named object from a module in the context of this function.
        """
        module = __import__(modulename, globals(), locals(), [name])
        return getattr(module, name)


    def parseXML(self, node):

        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_CALLBACK:
                self.callback = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_ARG:
                self.parseArgXML(child)


    def parseArgXML(self, node):
        name = None
        values = []
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_NAME:
                name = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_VALUE:
                if child.text is not None:
                    values.append(self.manager.server_info.subs(child.text.encode("utf-8")))
                else:
                    values.append("")
        if name:
            self.args[name] = values



class verify(object):
    """
    Defines how the result of a request should be verified. This is done
    by passing the response and response data to a callback with a set of arguments
    specified in the test XML config file. The callback name is in the XML config
    file also and is dynamically loaded to do the verification.
    """

    def __init__(self, manager):
        self.manager = manager
        self.require_features = set()
        self.exclude_features = set()
        self.callback = None
        self.args = {}


    def missingFeatures(self):
        return self.require_features - self.manager.server_info.features


    def excludedFeatures(self):
        return self.exclude_features & self.manager.server_info.features


    def doVerify(self, uri, response, respdata):

        # Re-do substitutions from values generated during the current test run
        if self.manager.server_info.hasextrasubs():
            for name, values in self.args.iteritems():
                newvalues = [self.manager.server_info.extrasubs(value) for value in values]
                self.args[name] = newvalues

        verifierClass = self._importName("verifiers." + self.callback, "Verifier")
        verifier = verifierClass()

        # Always clone the args as this verifier may be called multiple times
        args = dict((k, list(v)) for k, v in self.args.items())

        return verifier.verify(self.manager, uri, response, respdata, args)


    def _importName(self, modulename, name):
        """
        Import a named object from a module in the context of this function.
        """
        module = __import__(modulename, globals(), locals(), [name])
        return getattr(module, name)


    def parseXML(self, node):

        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_REQUIRE_FEATURE:
                self.parseFeatures(child, require=True)
            elif child.tag == src.xmlDefs.ELEMENT_EXCLUDE_FEATURE:
                self.parseFeatures(child, require=False)
            elif child.tag == src.xmlDefs.ELEMENT_CALLBACK:
                self.callback = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_ARG:
                self.parseArgXML(child)


    def parseFeatures(self, node, require=True):
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_FEATURE:
                (self.require_features if require else self.exclude_features).add(child.text.encode("utf-8"))


    def parseArgXML(self, node):
        name = None
        values = []
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_NAME:
                name = child.text.encode("utf-8")
            elif child.tag == src.xmlDefs.ELEMENT_VALUE:
                if child.text is not None:
                    values.append(self.manager.server_info.subs(child.text.encode("utf-8")))
                else:
                    values.append("")
        if name:
            self.args[name] = values



class stats(object):
    """
    Maintains stats about the current test.
    """

    def __init__(self):
        self.count = 0
        self.totaltime = 0.0
        self.currenttime = 0.0


    def startTimer(self):
        self.currenttime = time.time()


    def endTimer(self):
        self.count += 1
        self.totaltime += time.time() - self.currenttime



##
# Copyright (c) 2014-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from src.observers.log import Observer as LogObserver


class Observer(LogObserver):
    """
    A results observer that prints results to standard output.
    """

    _print_details = True

    def updateCalls(self):
        super(Observer, self).updateCalls()
        self._calls.update({
            "trace": self.trace,
        })


    def trace(self, text, indent=0):
        self.manager.logit(text)



##
# Copyright (c) 2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

def processHrefSubstitutions(hrefs, prefix):
    """
    Process the list of hrefs by prepending the supplied prefix. If the href is a
    list of hrefs, then prefix each item in the list and expand into the results. The
    empty string is represented by a single "-" in an href list.

    @param hrefs: list of URIs to process
    @type hrefs: L{list} of L{str}
    @param prefix: prefix to apply to each URI
    @type prefix: L{str}

    @return: resulting list of URIs
    @rtype: L{list} of L{str}
    """

    results = []
    for href in hrefs:
        if href.startswith("["):
            children = href[1:-1].split(",")
            results.extend([(prefix + (i if i != "-" else "")).rstrip("/") for i in children if i])
        else:
            results.append((prefix + href).rstrip("/"))

    return results



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Class to manage the testing process.
"""

from src.serverinfo import serverinfo
from xml.etree.cElementTree import ElementTree
from xml.parsers.expat import ExpatError
import getopt
import os
import random
import src.xmlDefs
import sys
import time

# Exceptions

EX_INVALID_CONFIG_FILE = "Invalid Config File"
EX_FAILED_REQUEST = "HTTP Request Failed"

class manager(object):

    """
    Main class that runs test suites defined in an XML config file.
    """

    RESULT_OK = 0
    RESULT_FAILED = 1
    RESULT_ERROR = 2
    RESULT_IGNORED = 3


    def __init__(self, text=True):
        self.server_info = serverinfo()
        self.data_dir = None
        self.pretest = None
        self.posttest = None
        self.tests = []
        self.textMode = text
        self.pid = 0
        self.memUsage = None
        self.randomSeed = None
        self.logFile = None
        self.digestCache = {}
        self.postgresLog = ""
        self.stoponfail = False
        self.print_request = False
        self.print_response = False
        self.print_request_response_on_error = False
        self.debug = False

        self.results = []
        self.totals = {
            self.RESULT_OK: 0,
            self.RESULT_FAILED: 0,
            self.RESULT_ERROR: 0,
            self.RESULT_IGNORED: 0
        }
        self.observers = []


    def logit(self, str):
        if self.logFile:
            self.logFile.write(str + "\n")
        print str


    def loadObserver(self, observer_name):
        module = __import__("observers." + observer_name, globals(), locals(), ["Observer", ])
        cl = getattr(module, "Observer")
        self.observers.append(cl(self))


    def message(self, message, *args, **kwargs):
        map(lambda x: x.message(message, *args, **kwargs), self.observers)


    def testProgress(self, count, total):
        results = {
            "count": count,
            "total": total,
        }
        self.message("testProgress", results)


    def testFile(self, name, details, result=None):
        self.results.append({
            "name": name,
            "details": details,
            "result": result,
            "tests": []
        })
        if result is not None:
            self.totals[result] += 1
        self.message("testFile", self.results[-1])
        return self.results[-1]["tests"]


    def testSuite(self, testfile, name, details, result=None):
        testfile.append({
            "name": name,
            "details": details,
            "result": result,
            "tests": []
        })
        if result is not None:
            self.totals[result] += 1
        self.message("testSuite", testfile[-1])
        return testfile[-1]["tests"]


    def testResult(self, testsuite, name, details, result, addons=None):
        result_details = {
            "name": name,
            "result": result,
            "details": details
        }
        if addons:
            result_details.update(addons)
        testsuite.append(result_details)
        self.totals[result] += 1
        self.message("testResult", testsuite[-1])


    def readXML(self, serverfile, testfiles, ssl, all, moresubs={}):

        self.message("trace", "Reading Server Info from \"{s}\"".format(s=serverfile))

        # Open and parse the server config file
        try:
            tree = ElementTree(file=serverfile)
        except ExpatError, e:
            raise RuntimeError("Unable to parse file '%s' because: %s" % (serverfile, e,))

        # Verify that top-level element is correct
        serverinfo_node = tree.getroot()
        if serverinfo_node.tag != src.xmlDefs.ELEMENT_SERVERINFO:
            raise EX_INVALID_CONFIG_FILE
        if not len(serverinfo_node):
            raise EX_INVALID_CONFIG_FILE
        self.server_info.parseXML(serverinfo_node)

        # Setup ssl stuff
        self.server_info.ssl = ssl
        self.server_info.port = self.server_info.sslport if ssl else self.server_info.nonsslport
        self.server_info.port2 = self.server_info.sslport2 if ssl else self.server_info.nonsslport2

        moresubs["$host:"] = "%s://%s" % (
            "https" if ssl else "http", self.server_info.host,
        )
        if (ssl and self.server_info.port != 443) or (not ssl and self.server_info.port != 80):
            moresubs["$host:"] += ":%d" % (self.server_info.port,)
        moresubs["$hostssl:"] = "https://%s" % (self.server_info.host,)
        if self.server_info.sslport != 443:
            moresubs["$hostssl:"] += ":%d" % (self.server_info.sslport,)

        moresubs["$host2:"] = "%s://%s" % (
            "https" if ssl else "http",
            self.server_info.host2,
        )
        if (ssl and self.server_info.port2 != 443) or (not ssl and self.server_info.port2 != 80):
            moresubs["$host2:"] += ":%d" % (self.server_info.port2,)
        moresubs["$hostssl2:"] = "https://%s" % (self.server_info.host2,)
        if self.server_info.sslport2 != 443:
            moresubs["$hostssl2:"] += ":%d" % (self.server_info.sslport2,)

        self.server_info.addsubs(moresubs)

        from src.caldavtest import caldavtest
        def _loadFile(fname, ignore_root=True):
            # Open and parse the config file
            try:
                tree = ElementTree(file=fname)
            except ExpatError, e:
                raise RuntimeError("Unable to parse file '%s' because: %s" % (fname, e,))
            caldavtest_node = tree.getroot()
            if caldavtest_node.tag != src.xmlDefs.ELEMENT_CALDAVTEST:
                if ignore_root:
                    self.message("trace", "Ignoring file \"{f}\" because it is not a test file".format(f=fname))
                    return None
                else:
                    raise EX_INVALID_CONFIG_FILE
            if not len(caldavtest_node):
                raise EX_INVALID_CONFIG_FILE

            self.message("Reading Test Details from \"{f}\"".format(f=fname))
            test = caldavtest(self, fname)
            test.parseXML(caldavtest_node)
            return test

        for ctr, testfile in enumerate(testfiles):
            self.message("load", testfile, ctr + 1, len(testfiles))

            # Open and parse the config file
            test = _loadFile(testfile)
            if test is None:
                continue

            # ignore if all mode and ignore-all is set
            if not all or not test.ignore_all:
                self.tests.append(test)

        if self.pretest is not None:
            self.pretest = _loadFile(self.pretest, False)
        if self.posttest is not None:
            self.posttest = _loadFile(self.posttest, False)

        self.message("load", None, ctr + 1, len(testfiles))


    def readCommandLine(self):
        sname = "scripts/server/serverinfo.xml"
        dname = "scripts/tests"
        basedir = None
        fnames = []
        ssl = False
        all = False
        excludes = set()
        subdir = None
        pidfile = "../CalendarServer/logs/caldavd.pid"
        random_order = False
        random_seed = str(random.randint(0, 1000000))
        observer_names = []

        options, args = getopt.getopt(
            sys.argv[1:],
            "s:mo:x:",
            [
                "ssl",
                "all",
                "basedir=",
                "subdir=",
                "exclude=",
                "pretest=",
                "posttest=",
                "observer=",
                "pid=",
                "postgres-log=",
                "random",
                "random-seed=",
                "stop",
                "print-details-onfail",
                "always-print-request",
                "always-print-response",
                "debug"
            ],
        )

        # Process single options
        for option, value in options:
            if option == "-s":
                sname = value
            elif option == "-x":
                dname = value
            elif option == "--ssl":
                ssl = True
            elif option == "--all":
                all = True
            elif option == "--basedir":
                basedir = value
                sname = os.path.join(basedir, "serverinfo.xml")
                dname = os.path.join(basedir, "tests")
                self.data_dir = os.path.join(basedir, "data")

                # Also add parent to PYTHON path
                sys.path.append(os.path.dirname(basedir))

            elif option == "--subdir":
                subdir = value + "/"
            elif option == "--exclude":
                excludes.add(value)
            elif option == "--pretest":
                self.pretest = value
            elif option == "--posttest":
                self.posttest = value
            elif option == "-m":
                self.memUsage = True
            elif option == "-o":
                self.logFile = open(value, "w")
            elif option == "--pid":
                pidfile = value
            elif option == "--observer":
                observer_names.append(value)
            elif option == "--postgres-log":
                self.postgresLog = value
            elif option == "--stop":
                self.stoponfail = True
            elif option == "--print-details-onfail":
                self.print_request_response_on_error = True
            elif option == "--always-print-request":
                self.print_request = True
            elif option == "--always-print-response":
                self.print_response = True
            elif option == "--random":
                random_order = True
            elif option == "--random-seed":
                random_seed = value
            elif option == "--debug":
                self.debug = True

        if all or not args:
            files = []
            os.path.walk(dname, lambda arg, dir, names: files.extend([os.path.join(dir, name) for name in names]) if not dir.startswith("test") else None, None)
            for file in files:
                if file.endswith(".xml") and file[len(dname) + 1:] not in excludes:
                    if subdir is None or file[len(dname) + 1:].startswith(subdir):
                        fnames.append(file)

        # Remove any server info file from files enumerated by --all
        fnames[:] = [x for x in fnames if (x != sname)]

        def _normPath(f):
            # paths starting with . or .. or /
            if f[0] in ('.', '/'):
                f = os.path.abspath(f)

                # remove unneeded leading path
                fsplit = f.split(dname)
                if 2 == len(fsplit):
                    f = dname + fsplit[1]

            # relative paths
            else:
                f = os.path.join(dname, f)
            return f

        # Process any file arguments as test configs
        for f in args:
            fnames.append(_normPath(f))

        if self.pretest is not None:
            self.pretest = _normPath(self.pretest)
        if self.posttest is not None:
            self.posttest = _normPath(self.posttest)

        # Randomize file list
        if random_order and len(fnames) > 1:
            random.seed(random_seed)
            random.shuffle(fnames)
            self.randomSeed = random_seed

        # Load observers
        map(lambda name: self.loadObserver(name), observer_names if observer_names else ["log", ])

        self.readXML(sname, fnames, ssl, all)

        if self.memUsage:
            fd = open(pidfile, "r")
            s = fd.read()
            self.pid = int(s)


    def runAll(self):

        startTime = time.time()

        self.message("start")

        ok = 0
        failed = 0
        ignored = 0
        try:
            for ctr, test in enumerate(self.tests):
                if len(self.tests) > 1:
                    self.testProgress(ctr + 1, len(self.tests))
                if self.pretest is not None:
                    o, f, i = self.pretest.run()

                    # Always stop the tests if the pretest fails
                    if f != 0:
                        break

                o, f, i = test.run()
                ok += o
                failed += f
                ignored += i

                if failed != 0 and self.stoponfail:
                    break

                if self.posttest is not None:
                    o, f, i = self.posttest.run()

                    # Always stop the tests if the posttest fails
                    if f != 0:
                        break

        except:
            failed += 1
            import traceback
            traceback.print_exc()

        endTime = time.time()

        self.timeDiff = endTime - startTime
        self.message("finish")

        if self.logFile is not None:
            self.logFile.close()

        return failed, endTime - startTime


    def getMemusage(self):
        """

        @param pid: numeric pid of process to get memory usage for
        @type pid:  int
        @retrun:    tuple of (RSS, VSZ) values for the process
        """

        fd = os.popen("ps -l -p %d" % (self.pid,))
        data = fd.read()
        lines = data.split("\n")
        procdata = lines[1].split()
        return int(procdata[6]), int(procdata[7])


    def getDataPath(self, fpath):
        return os.path.join(self.data_dir, fpath) if self.data_dir else fpath



#!/usr/bin/env python
#
##
# Copyright (c) 2013-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

import os
import datetime
import shutil
import sys
import argparse
from subprocess import Popen, PIPE

server_root = "/Applications/Server.app/Contents/ServerRoot"
os.environ["PATH"] = "%s/usr/bin:%s" % (server_root, os.environ["PATH"])
library_root = "/Library/Server/Calendar and Contacts"

directory_node = "/LDAPv3/127.0.0.1"


def cmd(args, input=None, raiseOnFail=True):

    if input:
        p = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
        result = p.communicate(input)
    else:
        p = Popen(args, stdout=PIPE, stderr=PIPE, shell=True)
        result = p.communicate()

    if raiseOnFail and p.returncode:
        raise RuntimeError(result[1])
    return result[0], p.returncode


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        description='Gather CalDAVTester diagnostics.',
    )
    parser.add_argument(
        '-d', '--directory', action='store',
        help='Destination directory for diagnostics archive'
    )
    args = parser.parse_args()

    print "Running CDT diagnostics due to test failure."
    log = []

    def error(message, e):
        log.append("CDT diagnostic: %s" % (message,))
        log.append(str(e))
        print "\n".join(log)
        sys.exit(1)

    now = datetime.datetime.now()
    now = now.replace(microsecond=0)
    dirname = "cdtdiagnose-%s" % (now.strftime("%Y%m%d-%H%M%S"),)

    if args.directory is not None:
        if not os.path.isdir(args.directory):
            print "Specified target directory path is invalid, using default."
        else:
            dirname = os.path.join(args.directory, dirname)

    print "Saving diagnostic archive to: {}".format(dirname,)
    try:
        os.mkdir(dirname)
    except Exception as e:
        error("Could not create archive directory: '%s'" % (dirname,), e)

    # Copy CDT log file file
    server_path = "cdt.txt"
    archive_path = os.path.join(dirname, os.path.basename(server_path))
    try:
        shutil.copy(server_path, archive_path)
    except Exception as e:
        error("Could not copy cdt results file: '%s' to '%s'" % (server_path, archive_path,), e)

    # Copy serverinfo file
    server_path = "scripts/server/serverinfo-caldav.xml"
    archive_path = os.path.join(dirname, os.path.basename(server_path))
    try:
        shutil.copy(server_path, archive_path)
    except Exception as e:
        error("Could not copy server info file: '%s' to '%s'" % (server_path, archive_path,), e)

    # Get server logs
    logs_path = os.path.join(library_root, "Logs")
    archive_path = os.path.join(dirname, "logs")
    try:
        shutil.copytree(logs_path, archive_path)
    except Exception as e:
        error("Could not copy server logs: '%s' to '%s'" % (logs_path, archive_path,), e)

    # Get server config files
    server_path = os.path.join(server_root, "etc", "caldavd")
    archive_path = os.path.join(dirname, "etc")
    try:
        shutil.copytree(server_path, archive_path)
    except Exception as e:
        error("Could not copy server conf: '%s' to '%s'" % (server_path, archive_path,), e)

    server_path = library_root
    archive_path = os.path.join(dirname, "Library")
    try:
        shutil.copytree(server_path, archive_path)
    except Exception as e:
        error("Could not copy library items: '%s' to '%s'" % (server_path, archive_path,), e)

    # Dump OD data
    try:
        results = ["*** Users"]
        results.extend(cmd("dscl %s -readall Users" % (directory_node,))[0].splitlines())
        results.append("\n\n*** Groups")
        results.extend(cmd("dscl %s -readall Groups" % (directory_node,))[0].splitlines())
        results.append("")

        with open(os.path.join(dirname, "dscl_dump.txt"), "w") as f:
            f.write("\n".join(results))
    except Exception as e:
        error("Could not dump OD data.", e)

    # Now archive the diagnostics data
    try:
        archive_name = shutil.make_archive(dirname, "gztar", dirname)
    except Exception as e:
        error("Could not make diagnostics archive.", e)

    print "Saved diagnostics to '%s'" % (archive_name,)



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Verifier that checks the response body for an exact match to data in a file.
"""

from difflib import unified_diff
from xml.etree.cElementTree import ElementTree, tostring
import StringIO
import os

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable
        # Get arguments
        files = args.get("filepath", [])
        if manager.data_dir:
            files = map(lambda x: os.path.join(manager.data_dir, x), files)
        filters = args.get("filter", [])

        # status code must be 200, 207
        if response.status not in (200, 207):
            return False, "        HTTP Status Code Wrong: %d" % (response.status,)

        # look for response data
        if not respdata:
            return False, "        No response body"

        # look for one file
        if len(files) != 1:
            return False, "        No file to compare response to"

        # read in all data from specified file
        fd = open(files[0], "r")
        try:
            try:
                data = fd.read()
            finally:
                fd.close()
        except:
            data = None

        if data is None:
            return False, "        Could not read data file"

        data = manager.server_info.subs(data)
        data = manager.server_info.extrasubs(data)

        def normalizeXMLData(data):
            # Read in XML
            try:
                tree = ElementTree(file=StringIO.StringIO(data))
            except Exception:
                raise ValueError("Could not parse XML data")

            # Apply filters
            for filter in filters:
                for node in tree.getiterator(filter):
                    node.clear()
            return tostring(tree.getroot())

        try:
            respdata = normalizeXMLData(respdata)
            data = normalizeXMLData(data)

            result = respdata == data

            if result:
                return True, ""
            else:
                error_diff = "\n".join([line for line in unified_diff(data.split("\n"), respdata.split("\n"))])
                return False, "        Response data does not exactly match file data %s" % (error_diff,)
        except Exception, e:
            return False, "        Response data is not xml data: %s" % (e,)



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Verifier that checks a propfind response to make sure that the specified ACL privileges
are available for the currently authenticated user.
"""

from xml.etree.cElementTree import ElementTree
from StringIO import StringIO
import urllib

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable

        granted = args.get("granted", [])
        denied = args.get("denied", [])

        # Process the multistatus response, extracting all current-user-privilege-set elements
        # and check to see that each required privilege is present, or that denied ones are not.

        # Must have MULTISTATUS response code
        if response.status != 207:
            return False, "           HTTP Status for Request: %d\n" % (response.status,)

        try:
            tree = ElementTree(file=StringIO(respdata))
        except Exception:
            return False, "           HTTP response is not valid XML: %d\n" % (respdata,)

        result = True
        resulttxt = ""
        for response in tree.findall("{DAV:}response"):

            # Get href for this response
            href = response.findall("{DAV:}href")
            if len(href) != 1:
                return False, "           Wrong number of DAV:href elements\n"
            href = urllib.unquote(href[0].text)

            # Get all privileges
            granted_privs = []
            privset = response.getiterator("{DAV:}current-user-privilege-set")
            for props in privset:
                # Determine status for this propstat
                privileges = props.findall("{DAV:}privilege")
                for privilege in privileges:
                    for child in privilege.getchildren():
                        granted_privs.append(child.tag)

            granted_result_set = set(granted_privs)
            granted_test_set = set(granted)
            denied_test_set = set(denied)

            # Now do set difference
            granted_missing = granted_test_set.difference(granted_result_set)
            denied_present = granted_result_set.intersection(denied_test_set)

            if len(granted_missing) + len(denied_present) != 0:
                if len(granted_missing) != 0:
                    l = list(granted_missing)
                    resulttxt += "        Missing privileges not granted for %s:" % href
                    for i in l:
                        resulttxt += " " + str(i)
                    resulttxt += "\n"
                if len(denied_present) != 0:
                    l = list(denied_present)
                    resulttxt += "        Available privileges that should be denied for %s:" % href
                    for i in l:
                        resulttxt += " " + str(i)
                    resulttxt += "\n"
                result = False

        return result, resulttxt



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from difflib import unified_diff
from pycalendar.vcard.card import Card
import os

"""
Verifier that checks the response body for a semantic match to data in a file.
"""

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args, is_json=False): #@UnusedVariable
        # Get arguments
        files = args.get("filepath", [])
        if manager.data_dir:
            files = map(lambda x: os.path.join(manager.data_dir, x), files)
        carddata = args.get("data", [])
        filters = args.get("filter", [])

        # Always remove these
        filters.append("PRODID")
        filters.append("REV")

        # status code must be 200, 201, 207
        if response.status not in (200, 201, 207):
            return False, "        HTTP Status Code Wrong: %d" % (response.status,)

        # look for response data
        if not respdata:
            return False, "        No response body"

        # look for one file
        if len(files) != 1 and len(carddata) != 1:
            return False, "        No file to compare response to"

        # read in all data from specified file or use provided data
        if len(files):
            fd = open(files[0], "r")
            try:
                try:
                    data = fd.read()
                finally:
                    fd.close()
            except:
                data = None
        else:
            data = carddata[0] if len(carddata) else None

        if data is None:
            return False, "        Could not read data file"

        data = manager.server_info.subs(data)

        def removePropertiesParameters(component):

            allProps = []
            for properties in component.getProperties().itervalues():
                allProps.extend(properties)
            for property in allProps:
                for filter in filters:
                    if ":" in filter:
                        propname, parameter = filter.split(":")
                        if property.getName() == propname:
                            if property.hasParameter(parameter):
                                property.removeParameters(parameter)
                    else:
                        if property.getName() == filter:
                            component.removeProperty(property)

        try:
            format = Card.sFormatJSON if is_json else Card.sFormatText

            resp_adbk = Card.parseData(respdata, format=format)
            removePropertiesParameters(resp_adbk)
            respdata = resp_adbk.getText(format=format)

            data_adbk = Card.parseData(data, format=format)
            removePropertiesParameters(data_adbk)
            data = data_adbk.getText(format=format)

            result = resp_adbk == data_adbk

            if result:
                return True, ""
            else:
                error_diff = "\n".join([line for line in unified_diff(data.split("\n"), respdata.split("\n"))])
                return False, "        Response data does not exactly match file data%s" % (error_diff,)
        except Exception, e:
            return False, "        Response data is not address data: %s" % (e,)



##
# Copyright (c) 2010-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##


"""
Verifier that matches JSON content using extended JSON pointer syntax.

JSON pointer syntax is extended as follows:

1) A ~$xxx at the end will result in a test for the string "xxx" in the matching JSON object.
2) A "." as a path segment will match any JSON object member or array item.
"""

import json
from src.jsonPointer import JSONMatcher, JSONPointerMatchError

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable
        # Get arguments
        statusCodes = args.get("status", ["200", ])
        exists = args.get("exists", [])
        notexists = args.get("notexists", [])

        # status code must match
        if str(response.status) not in statusCodes:
            return False, "        HTTP Status Code Wrong: %d" % (response.status,)

        # look for response data
        if not respdata:
            return False, "        No response body"

        # Must be application/json
        ct = response.msg.getheaders("content-type")[0].split(";")[0]
        if ct != "application/json" and not ct.endswith("+json"):
            return False, "        Wrong Content-Type: %s" % (ct,)

        # Read in json
        try:
            j = json.loads(respdata)
        except Exception, e:
            return False, "        Response data is not JSON data: %s" % (e,)

        def _splitPathTests(path):
            if '[' in path:
                return path.split('[', 1)
            else:
                return path, None

        result = True
        resulttxt = ""
        for jpath in exists:
            if jpath.find("~$") != -1:
                path, value = jpath.split("~$")
            else:
                path, value = jpath, None
            try:
                jp = JSONMatcher(path)
            except Exception:
                result = False
                resulttxt += "        Invalid JSON pointer for %s\n" % (path,)
            else:
                try:
                    jobjs = jp.match(j)
                    if not jobjs:
                        result = False
                        resulttxt += "        Items not returned in JSON for %s\n" % (path,)
                    if value and value not in map(str, jobjs):
                        result = False
                        resulttxt += "        Item values not returned in JSON for %s\n" % (jpath,)
                except JSONPointerMatchError:
                    result = False
                    resulttxt += "        Items not returned in JSON for %s\n" % (path,)

        for jpath in notexists:
            if jpath.find("~$") != -1:
                path, value = jpath.split("~$")
            else:
                path, value = jpath, None
            try:
                jp = JSONMatcher(path)
            except Exception:
                result = False
                resulttxt += "        Invalid JSON pointer for %s\n" % (jpath,)
            else:
                try:
                    jobjs = jp.match(j)
                except JSONPointerMatchError:
                    pass
                else:
                    if len(jobjs):
                        resulttxt += "        Items returned in JSON for %s\n" % (jpath,)
                        result = False

        return result, resulttxt



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Verifier that checks the response headers for a specific value.
"""

import re

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable
        # Split into header/value tuples
        testheader = args.get("header", [])[:]
        for i in range(len(testheader)):
            p = testheader[i]
            present = "single"
            if p[0] == "!":
                p = p[1:]
                present = "none"
            if p[0] == "*":
                p = p[1:]
                present = "multiple"
            if p.find("$") != -1:
                testheader[i] = (p.split("$", 1)[0], p.split("$", 1)[1], present, True,)
            elif p.find("!") != -1:
                testheader[i] = (p.split("!", 1)[0], p.split("!", 1)[1], present, False,)
            else:
                testheader[i] = (p, None, present, True,)

        result = True
        resulttxt = ""
        for hdrname, hdrvalue, presence, matchvalue in testheader:
            hdrs = response.msg.getheaders(hdrname)
            if (hdrs is None or (len(hdrs) == 0)):
                if presence != "none":
                    result = False
                    if len(resulttxt):
                        resulttxt += "\n"
                    resulttxt += "        Missing Response Header: %s" % (hdrname,)
                    continue
                else:
                    continue

            if (hdrs is not None) and (len(hdrs) != 0) and (presence == "none"):
                result = False
                if len(resulttxt):
                    resulttxt += "\n"
                resulttxt += "        Response Header was present one or more times: %s" % (hdrname,)
                continue

            if (len(hdrs) != 1) and (presence == "single"):
                result = False
                if len(resulttxt):
                    resulttxt += "\n"
                resulttxt += "        Multiple Response Headers: %s" % (hdrname,)
                continue

            if (hdrvalue is not None):
                hdrvalue = hdrvalue.replace(" ", "")
                matched = False
                for hdr in hdrs:
                    hdr = hdr.replace(" ", "")
                    if (re.match(hdrvalue, hdr) is not None):
                        matched = True
                        break

                if matchvalue and not matched or not matchvalue and matched:
                    result = False
                    if len(resulttxt):
                        resulttxt += "\n"
                    resulttxt += "        Wrong Response Header Value: %s: %s" % (hdrname, str(hdrs))

        return result, resulttxt



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from difflib import unified_diff
from pycalendar.icalendar.calendar import Calendar
from pycalendar.icalendar.componentrecur import ComponentRecur
from pycalendar.parameter import Parameter
import os

"""
Verifier that checks the response body for a semantic match to data in a file.
"""

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args, is_json=False): #@UnusedVariable
        # Get arguments
        files = args.get("filepath", [])
        if manager.data_dir:
            files = map(lambda x: os.path.join(manager.data_dir, x), files)
        caldata = args.get("data", [])
        filters = args.get("filter", [])
        statusCode = args.get("status", ["200", "201", "207"])
        doTimezones = args.get("doTimezones", None)

        if "EMAIL parameter" not in manager.server_info.features:
            filters.append("ATTENDEE:EMAIL")
            filters.append("ORGANIZER:EMAIL")
        filters.append("ATTENDEE:X-CALENDARSERVER-DTSTAMP")
        filters.append("ATTENDEE:X-CALENDARSERVER-AUTO")
        filters.append("ATTENDEE:X-CALENDARSERVER-RESET-PARTSTAT")
        filters.append("CALSCALE")
        filters.append("PRODID")
        filters.append("DTSTAMP")
        filters.append("CREATED")
        filters.append("LAST-MODIFIED")
        filters.append("X-WR-CALNAME")

        for afilter in tuple(filters):
            if afilter[0] == "!" and afilter[1:] in filters:
                filters.remove(afilter[1:])
        filters = filter(lambda x: x[0] != "!", filters)

        if doTimezones is None:
            doTimezones = "timezones-by-reference" not in manager.server_info.features
        else:
            doTimezones = doTimezones == "true"

        # status code must be 200, 201, 207 or explicitly specified code
        if str(response.status) not in statusCode:
            return False, "        HTTP Status Code Wrong: %d" % (response.status,)

        # look for response data
        if not respdata:
            return False, "        No response body"

        # look for one file
        if len(files) != 1 and len(caldata) != 1:
            return False, "        No file to compare response to"

        # read in all data from specified file or use provided data
        if len(files):
            fd = open(files[0], "r")
            try:
                try:
                    data = fd.read()
                finally:
                    fd.close()
            except:
                data = None
        else:
            data = caldata[0] if len(caldata) else None

        if data is None:
            return False, "        Could not read data file"

        data = manager.server_info.extrasubs(manager.server_info.subs(data))

        def removePropertiesParameters(component):

            if not doTimezones:
                for subcomponent in tuple(component.getComponents()):
                    if subcomponent.getType() == "VTIMEZONE":
                        component.removeComponent(subcomponent)

            for subcomponent in component.getComponents():
                removePropertiesParameters(subcomponent)

            allProps = []
            for properties in component.getProperties().itervalues():
                allProps.extend(properties)
            for property in allProps:
                # Always reset DTSTAMP on these properties
                if property.getName() in ("ATTENDEE", "X-CALENDARSERVER-ATTENDEE-COMMENT"):
                    if property.hasParameter("X-CALENDARSERVER-DTSTAMP"):
                        property.replaceParameter(Parameter("X-CALENDARSERVER-DTSTAMP", "20080101T000000Z"))

                for filter in filters:
                    if ":" in filter:
                        propname, parameter = filter.split(":")
                        if property.getName() == propname:
                            if property.hasParameter(parameter):
                                property.removeParameters(parameter)
                    else:
                        if property.getName() == filter:
                            component.removeProperty(property)

        def reconcileRecurrenceOverrides(calendar1, calendar2):
            """
            Make sure that the same set of overridden components appears in both calendar objects.
            """
            def _getRids(calendar):
                """
                Get all the recurrence ids of the specified calendar.
                """
                results = set()
                master = None
                for subcomponent in calendar.getComponents():
                    if isinstance(subcomponent, ComponentRecur):
                        rid = subcomponent.getRecurrenceID()
                        if rid:
                            results.add(rid.duplicateAsUTC())
                        else:
                            master = subcomponent
                return results, master

            def _addOverrides(calendar, master, missing_rids):
                """
                Derive instances for the missing overrides in the specified calendar object.
                """
                if master is None or not missing_rids:
                    return
                for rid in missing_rids:
                    # If we were fed an already derived component, use that, otherwise make a new one
                    newcomp = calendar.deriveComponent(rid)
                    if newcomp is not None:
                        calendar.addComponent(newcomp)

            rids1, master1 = _getRids(calendar1)
            rids2, master2 = _getRids(calendar2)

            _addOverrides(calendar1, master1, rids2 - rids1)
            _addOverrides(calendar2, master2, rids1 - rids2)

        try:
            format = Calendar.sFormatJSON if is_json else Calendar.sFormatText

            resp_calendar = Calendar.parseData(respdata, format=format)
            removePropertiesParameters(resp_calendar)

            data_calendar = Calendar.parseData(data, format=format)
            removePropertiesParameters(data_calendar)

            reconcileRecurrenceOverrides(resp_calendar, data_calendar)

            respdata = resp_calendar.getText(includeTimezones=Calendar.NO_TIMEZONES, format=format)
            data = data_calendar.getText(includeTimezones=Calendar.NO_TIMEZONES, format=format)

            result = resp_calendar == data_calendar

            if result:
                return True, ""
            else:
                error_diff = "\n".join([line for line in unified_diff(data.split("\n"), respdata.split("\n"))])
                return False, "        Response data does not exactly match file data%s" % (error_diff,)
        except Exception, e:
            return False, "        Response data is not calendar data: %s" % (e,)



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Verifier that checks the response of a free-busy-query.
"""

from pycalendar.icalendar.calendar import Calendar
from pycalendar.exceptions import InvalidData

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable

        # Must have status 200
        if response.status != 200:
            return False, "        HTTP Status Code Wrong: %d" % (response.status,)

        # Get expected FREEBUSY info
        busy = args.get("busy", [])
        tentative = args.get("tentative", [])
        unavailable = args.get("unavailable", [])

        # Parse data as calendar object
        try:
            calendar = Calendar.parseText(respdata)

            # Check for calendar
            if calendar is None:
                raise ValueError("Not a calendar: %s" % (respdata,))

            # Only one component
            comps = calendar.getComponents("VFREEBUSY")
            if len(comps) != 1:
                raise ValueError("Wrong number or unexpected components in calendar")

            # Must be VFREEBUSY
            fb = comps[0]

            # Extract periods
            busyp = []
            tentativep = []
            unavailablep = []
            for fp in fb.getProperties("FREEBUSY"):
                periods = fp.getValue().getValues()
                # Convert start/duration to start/end
                for i in range(len(periods)):
                    periods[i].getValue().setUseDuration(False)
                # Check param
                fbtype = "BUSY"
                if fp.hasParameter("FBTYPE"):
                    fbtype = fp.getParameterValue("FBTYPE")
                if fbtype == "BUSY":
                    busyp.extend(periods)
                elif fbtype == "BUSY-TENTATIVE":
                    tentativep.extend(periods)
                elif fbtype == "BUSY-UNAVAILABLE":
                    unavailablep.extend(periods)
                else:
                    raise ValueError("Unknown FBTYPE: %s" % (fbtype,))

            # Set sizes must match
            if (
                (len(busy) != len(busyp)) or
                (len(unavailable) != len(unavailablep)) or
                (len(tentative) != len(tentativep))
            ):
                raise ValueError("Period list sizes do not match.")

            # Convert to string sets
            busy = set(busy)
            busyp = [x.getValue().getText() for x in busyp]
            busyp = set(busyp)
            tentative = set(tentative)
            tentativep = [x.getValue().getText() for x in tentativep]
            tentativep = set(tentativep)
            unavailable = set(unavailable)
            unavailablep = [x.getValue().getText() for x in unavailablep]
            unavailablep = set(unavailablep)

            # Compare all periods
            if len(busyp.symmetric_difference(busy)):
                raise ValueError("Busy periods do not match")
            elif len(tentativep.symmetric_difference(tentative)):
                raise ValueError("Busy-tentative periods do not match")
            elif len(unavailablep.symmetric_difference(unavailable)):
                raise ValueError("Busy-unavailable periods do not match")

        except InvalidData:
            return False, "        HTTP response data is not a calendar"
        except ValueError, txt:
            return False, "        HTTP response data is invalid: %s" % (txt,)

        return True, ""



##
# Copyright (c) 2014-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

from src.observers.base import BaseResultsObserver

class Observer(BaseResultsObserver):
    """
    A results observer that prints when a test file is loaded.
    """

    def updateCalls(self):
        super(Observer, self).updateCalls()
        self._calls.update({
            "load": self.load,
        })


    def load(self, name, current, total):
        """
        Message triggered when loading a script file.

        @param name: name of file being loaded, or L{None} for last file
        @type name: L{str}
        @param current: current number of files loaded
        @type current: L{int}
        @param total: total number of files to load
        @type total: L{int}
        """

        if name is not None:
            self.manager.logit("Loading {current} of {total}: {name}".format(current=current, total=total, name=name))
        else:
            self.manager.logit("Loading files complete.\n")



##
# Copyright (c) 2010-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##


"""
Verifier that checks the response body for an exact match to data in a file.
"""

from pycalendar.icalendar.calendar import Calendar
from xml.etree.cElementTree import ElementTree
import json
import re
import StringIO

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable
        # Get arguments
        parent = args.get("parent", [])
        exists = args.get("exists", [])
        notexists = args.get("notexists", [])

        # status code must be 200, 207
        if response.status not in (200, 207):
            return False, "        HTTP Status Code Wrong: %d" % (response.status,)

        # look for response data
        if not respdata:
            return False, "        No response body"

        # Read in XML
        try:
            tree = ElementTree(file=StringIO.StringIO(respdata))
        except Exception, e:
            return False, "        Response data is not xml data: %s" % (e,)

        def _splitPathTests(path):
            if '[' in path:
                return path.split('[', 1)
            else:
                return path, None

        if parent:
            nodes = self.nodeForPath(tree.getroot(), parent[0])
            if len(nodes) == 0:
                return False, "        Response data is missing parent node: %s" % (parent[0],)
            elif len(nodes) > 1:
                return False, "        Response data has too many parent nodes: %s" % (parent[0],)
            root = nodes[0]
        else:
            root = tree.getroot()

        result = True
        resulttxt = ""
        for path in exists:

            matched, txt = self.matchNode(root, path)
            result &= matched
            resulttxt += txt

        for path in notexists:
            matched, _ignore_txt = self.matchNode(root, path)
            if matched:
                resulttxt += "        Items returned in XML for %s\n" % (path,)
                result = False

        return result, resulttxt


    def nodeForPath(self, root, path):
        if '[' in path:
            actual_path, tests = path.split('[', 1)
        else:
            actual_path = path
            tests = None

        # Handle absolute root element
        if actual_path[0] == '/':
            actual_path = actual_path[1:]
        r = re.search("(\{[^\}]+\}[^/]+)(.*)", actual_path)
        if r.group(2):
            root_path = r.group(1)
            child_path = r.group(2)[1:]
            if root.tag != root_path:
                return None
            nodes = root.findall(child_path)
        else:
            root_path = actual_path
            child_path = None
            nodes = (root,)

        if len(nodes) == 0:
            return None

        results = []

        if tests:
            tests = [item[:-1] for item in tests.split('[')]
            for test in tests:
                for node in nodes:
                    testresult = self.testNode(node, path, test)
                    if testresult is None:
                        results.append(node)
        else:
            results = nodes

        return results


    @classmethod
    def testNode(cls, node, node_path, test):
        result = None
        if test[0] == '@':
            if '=' in test:
                attr, value = test[1:].split('=')
                value = value[1:-1]
            else:
                attr = test[1:]
                value = None
            if attr not in node.keys():
                result = "        Missing attribute returned in XML for %s\n" % (node_path,)
            if value is not None and node.get(attr) != value:
                result = "        Incorrect attribute value returned in XML for %s\n" % (node_path,)
        elif test[0] == '=':
            if node.text != test[1:]:
                result = "        Incorrect value returned in XML for %s\n" % (node_path,)
        elif test[0] == '!':
            if node.text == test[1:]:
                result = "        Incorrect value returned in XML for %s\n" % (node_path,)
        elif test[0] == '*':
            if node.text is None or node.text.find(test[1:]) == -1:
                result = "        Incorrect value returned in XML for %s\n" % (node_path,)
        elif test[0] == '$':
            if node.text is None or node.text.find(test[1:]) != -1:
                result = "        Incorrect value returned in XML for %s\n" % (node_path,)
        elif test[0] == '+':
            if node.text is None or not node.text.startswith(test[1:]):
                result = "        Incorrect value returned in XML for %s\n" % (node_path,)
        elif test[0] == '^':
            if "=" in test:
                element, value = test[1:].split("=", 1)
            else:
                element = test[1:]
                value = None
            for child in node.getchildren():
                if child.tag == element and (value is None or child.text == value):
                    break
            else:
                result = "        Missing child returned in XML for %s\n" % (node_path,)
        elif test[0] == '|':
            if len(test) == 2 and test[1] == "|":
                if node.text is None and len(node.getchildren()) == 0:
                    result = "        Empty element returned in XML for %s\n" % (node_path,)
            else:
                if node.text is not None or len(node.getchildren()) != 0:
                    result = "        Non-empty element returned in XML for %s\n" % (node_path,)

        # Try to parse as iCalendar
        elif test == 'icalendar':
            try:
                Calendar.parseText(node.text)
            except:
                result = "        Incorrect value returned in iCalendar for %s\n" % (node_path,)

        # Try to parse as JSON
        elif test == 'json':
            try:
                json.loads(node.text)
            except:
                result = "        Incorrect value returned in XML for %s\n" % (node_path,)
        return result


    @classmethod
    def matchNode(cls, root, xpath, parent_map=None, title=None):

        if title is None:
            title = xpath
        result = True
        resulttxt = ""

        # Find the first test in the xpath
        if '[' in xpath:
            actual_xpath, tests = xpath.split('[', 1)
        else:
            actual_xpath = xpath
            tests = None

        if parent_map is None:
            parent_map = dict((c, p) for p in root.getiterator() for c in p)

        # Handle parents
        if actual_xpath.startswith("../"):
            root = parent_map[root]
            actual_xpath = "./" + actual_xpath[3:]

        # Handle absolute root element and find all matching nodes
        r = re.search("(/?\{[^\}]+\}[^/]+|\.)(.*)", actual_xpath)
        if r.group(2):
            root_path = r.group(1)
            child_path = r.group(2)[1:]
            if root_path != "." and root.tag != root_path.lstrip("/"):
                resulttxt += "        Items not returned in XML for %s\n" % (title,)
                result = False
                return result, resulttxt
            nodes = root.findall(child_path)
        else:
            nodes = (root,)

        if len(nodes) == 0:
            resulttxt += "        Items not returned in XML for %s\n" % (title,)
            result = False
            return result, resulttxt


        if tests:
            # Split the tests into tests plus additional path
            pos = tests.find("]/")
            if pos != -1:
                node_tests = tests[:pos + 1]
                next_path = tests[pos + 1:]
            else:
                node_tests = tests
                next_path = None

            node_tests = [item[:-1] for item in node_tests.split('[')]
            for test in node_tests:
                for node in nodes:
                    testresult = cls.testNode(node, title, test)
                    if testresult is None:
                        if next_path:
                            next_result, testresult = cls.matchNode(node, next_path[1:], parent_map, title)
                            if next_result:
                                break
                        else:
                            break

                if testresult:
                    resulttxt += testresult
                    result = False
                    break

        return result, resulttxt


# Tests
if __name__ == '__main__':
    xmldata = """
<D:test xmlns:D="DAV:">
    <D:a>A</D:a>
    <D:b>
        <D:c>C</D:c>
        <D:d>D</D:d>
    </D:b>
    <D:b>
        <D:c>C</D:c>
        <D:d>F</D:d>
    </D:b>
</D:test>
"""

    node = ElementTree(file=StringIO.StringIO(xmldata)).getroot()

    assert Verifier.matchNode(node, "/{DAV:}test/{DAV:}b/{DAV:}c[=C]/../{DAV:}d[=D]")[0]
    assert not Verifier.matchNode(node, "/{DAV:}test/{DAV:}b/{DAV:}c[=C]/../{DAV:}d[=E]")[0]



##
# Copyright (c) 2014-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##
from src.manager import manager
from src.observers.base import BaseResultsObserver


class Observer(BaseResultsObserver):
    """
    A results observer that prints results to standard output.
    """

    RESULT_STRINGS = {
        manager.RESULT_OK: "[OK]",
        manager.RESULT_FAILED: "[FAILED]",
        manager.RESULT_ERROR: "[ERROR]",
        manager.RESULT_IGNORED: "[IGNORED]",
    }

    _print_details = False

    def __init__(self, manager):
        super(Observer, self).__init__(manager)
        self.loggedFailures = []
        self.currentFile = None
        self.currentSuite = None


    def updateCalls(self):
        super(Observer, self).updateCalls()
        self._calls.update({
            "start": self.start,
            "testProgress": self.testProgress,
            "testFile": self.testFile,
            "testSuite": self.testSuite,
            "testResult": self.testResult,
            "finish": self.finish,
        })


    def start(self):
        self.manager.logit("Starting tests")
        if self.manager.randomSeed is not None:
            self.manager.logit("Randomizing order using seed '{rs}'".format(rs=self.manager.randomSeed))


    def testProgress(self, result):
        self.manager.logit("")
        self.manager.logit("File {count} of {total}".format(**result))


    def testFile(self, result):
        self.currentFile = result["name"].replace("/", ".")[:-4]
        self.manager.logit("")
        self._logResult(self.currentFile, result)
        if result["result"] in (manager.RESULT_FAILED, manager.RESULT_ERROR):
            failtxt = "{result}\n{details}\n\n{file}".format(
                result=self.RESULT_STRINGS[result["result"]],
                details=result["details"],
                file=self.currentFile,
            )
            self.loggedFailures.append(failtxt)


    def testSuite(self, result):
        self.currentSuite = result["name"]
        result_name = "  Suite: " + result["name"]
        self._logResult(result_name, result)
        if result["result"] in (manager.RESULT_FAILED, manager.RESULT_ERROR):
            failtxt = "{result}\n{details}\n\n{file}/{suite}".format(
                result=self.RESULT_STRINGS[result["result"]],
                details=result["details"],
                file=self.currentFile,
                suite=self.currentSuite,
            )
            self.loggedFailures.append(failtxt)


    def testResult(self, result):
        result_name = "    Test: " + result["name"]
        self._logResult(result_name, result)
        if result["result"] in (manager.RESULT_FAILED, manager.RESULT_ERROR):
            failtxt = "{result}\n{details}\n\n{file}/{suite}/{test}".format(
                result=self.RESULT_STRINGS[result["result"]],
                details=result["details"],
                file=self.currentFile,
                suite=self.currentSuite,
                test=result["name"],
            )
            self.loggedFailures.append(failtxt)


    def _logResult(self, name, result):
        if result["result"] is not None:
            result_value = self.RESULT_STRINGS[result["result"]]
            self.manager.logit("{name:<60}{value:>10}".format(name=name, value=result_value))
        else:
            self.manager.logit("{name:<60}".format(name=name))
        if self._print_details and result["details"]:
            self.manager.logit(result["details"])


    def finish(self):
        self.manager.logit("")
        if self.manager.totals[manager.RESULT_FAILED] + self.manager.totals[manager.RESULT_ERROR] != 0:
            for failed in self.loggedFailures:
                self.manager.logit("=" * 70)
                self.manager.logit(failed)
            overall = "FAILED (ok={o}, ignored={i}, failed={f}, errors={e})".format(
                o=self.manager.totals[manager.RESULT_OK],
                i=self.manager.totals[manager.RESULT_IGNORED],
                f=self.manager.totals[manager.RESULT_FAILED],
                e=self.manager.totals[manager.RESULT_ERROR],
            )
        else:
            overall = "PASSED (ok={o}, ignored={i})".format(
                o=self.manager.totals[manager.RESULT_OK],
                i=self.manager.totals[manager.RESULT_IGNORED],
            )
        self.manager.logit("-" * 70)
        self.manager.logit("Ran {total} tests in {time:.3f}s\n".format(
            total=sum(self.manager.totals.values()),
            time=self.manager.timeDiff,
        ))

        self.manager.logit(overall)



#!/usr/bin/env python
#
##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##
#
# Runs the CalDAVTester test suite ensuring that required packages are available.
#

import getopt
import os
import subprocess
import sys

cwd = os.getcwd()
top = cwd[:cwd.rfind("/")]
add_paths = []
svn = "/usr/bin/svn"
uri_base = "http://svn.calendarserver.org/repository/calendarserver"

packages = [
    ("pycalendar", "pycalendar/src", uri_base + "/PyCalendar/trunk", "HEAD"),
]

def usage():
    print """Usage: run.py [options]
Options:
    -h       Print this help and exit
    -s       Do setup only - do not run any tests
    -r       Run tests only - do not do setup
    -p       Print PYTHONPATH
"""



def setup():
    for package in packages:
        ppath = "%s/%s" % (top, package[0],)
        if not os.path.exists(ppath):
            print "%s package is not present." % (package[0],)
            os.system("%s checkout -r %s %s@%s %s" % (svn, package[3], package[2], package[3], ppath,))
        else:
            print "%s package is present." % (package[0],)
            fd = os.popen("%s info ../%s --xml" % (svn, package[0],))
            line = fd.read()
            wc_url = line[line.find("<url>") + 5:line.find("</url>")]
            if wc_url != package[2]:
                print "Current working copy (%s) is from the wrong URI: %s != %s, switching..." % (ppath, wc_url, package[2],)
                os.system("%s switch -r %s %s %s" % (svn, package[3], package[2], ppath,))
            else:
                rev = line[line.find("revision=\"") + 10:]
                rev = rev[:rev.find("\"")]
                if rev != package[3]:
                    print "Updating %s..." % (package[0],)
                    os.system("%s update -r %s %s" % (svn, package[3], ppath,))

        add_paths.append("%s/%s" % (top, package[1],))



def pythonpath():
    for package in packages:
        add_paths.append("%s/%s" % (top, package[1],))
    pypaths = sys.path
    pypaths.extend(add_paths)
    return ":".join(pypaths)



def runit():
    pythonpath = ":".join(add_paths)
    return subprocess.Popen(["./testcaldav.py", "--all"], env={"PYTHONPATH": pythonpath}).wait()



if __name__ == "__main__":

    try:
        do_setup = True
        do_run = True

        options, args = getopt.getopt(sys.argv[1:], "hprs")

        for option, value in options:
            if option == "-h":
                usage()
                sys.exit(0)
            elif option == "-p":
                print pythonpath()
                sys.exit(0)
            elif option == "-r":
                do_setup = False
            elif option == "-s":
                do_run = False
            else:
                print "Unrecognized option: %s" % (option,)
                usage()
                raise ValueError

        # Process arguments
        if len(args) != 0:
            print "No arguments allowed."
            usage()
            raise ValueError

        if (do_setup):
            setup()
        else:
            pythonpath()
        if (do_run):
            sys.exit(runit())
        else:
            sys.exit(0)
    except SystemExit, e:
        pass
    except Exception, e:
        sys.exit(str(e))



##
# Copyright (c) 2014-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##


class BaseResultsObserver(object):
    """
    A base class for an observer that gets passed results of tests.

    Supported messages:

    trace - tracing tool activity
    begin - beginning
    load - loading a test
    start - starting tests
    testFile - add a test file
    testSuite - add a test suite
    testResult - add a test result
    finish - tests completed
    """

    def __init__(self, manager):
        self.updateCalls()
        self.manager = manager


    def updateCalls(self):
        self._calls = {}


    def message(self, message, *args, **kwargs):

        callit = self._calls.get(message)

        if callit is not None:
            callit(*args, **kwargs)



#!/usr/bin/env python
#
##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##
#
# Runs a suite of tests defined in a config file against a caldav server
#

from src.manager import manager
import sys

if __name__ == "__main__":

    manager = manager()
    manager.readCommandLine()
    result, timing = manager.runAll()
    sys.exit(result)



#!/usr/bin/env python
# coding=utf-8
#
##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##
#
# Creates some test accounts on an OpenDirectory server for use with CalDAVTester
#

from getpass import getpass
from plistlib import readPlistFromString
from plistlib import writePlist
from subprocess import Popen, PIPE
import getopt
import os
import sys
import traceback
import uuid
import xml.parsers.expat

"""
OpenDirectory.framework
"""

import objc as _objc

__bundle__ = _objc.initFrameworkWrapper(
    "OpenDirectory",
    frameworkIdentifier="com.apple.OpenDirectory",
    frameworkPath=_objc.pathForFramework(
        "/System/Library/Frameworks/OpenDirectory.framework"
    ),
    globals=globals()
)

# DS attributes we need
kDSStdRecordTypeUsers = "dsRecTypeStandard:Users"
kDSStdRecordTypeGroups = "dsRecTypeStandard:Groups"
kDSStdRecordTypePlaces = "dsRecTypeStandard:Places"
kDSStdRecordTypeResources = "dsRecTypeStandard:Resources"

kDS1AttrGeneratedUID = "dsAttrTypeStandard:GeneratedUID"
kDSNAttrRecordName = "dsAttrTypeStandard:RecordName"

eDSExact = 0x2001


sys_root = "/Applications/Server.app/Contents/ServerRoot"
os.environ["PATH"] = "%s/usr/bin:%s" % (sys_root, os.environ["PATH"])

diradmin_user = "admin"
diradmin_pswd = ""
directory_node = "/LDAPv3/127.0.0.1"
utility = sys_root + "/usr/sbin/calendarserver_manage_principals"
cmdutility = sys_root + "/usr/sbin/calendarserver_command_gateway"
configutility = sys_root + "/usr/sbin/calendarserver_config"

verbose = False
veryverbose = False

serverinfo_template = "scripts/server/serverinfo-template.xml"

details = {
    "caldav": {
        "serverinfo": "scripts/server/serverinfo-caldav.xml"
    },
}

base_dir = "../CalendarServer/"

number_of_users = 40
number_of_groups = 40
number_of_publics = 10
number_of_resources = 20
number_of_locations = 10

guids = {
    "testadmin"  : "",
    "apprentice" : "",
    "i18nuser"   : "",
}

for i in range(1, number_of_users + 1):
    guids["user%02d" % (i,)] = ""

for i in range(1, number_of_publics + 1):
    guids["public%02d" % (i,)] = ""

for i in range(1, number_of_groups + 1):
    guids["group%02d" % (i,)] = ""

for i in range(1, number_of_resources + 1):
    guids["resource%02d" % (i,)] = ""

for i in range(1, number_of_locations + 1):
    guids["location%02d" % (i,)] = ""

locations = {}
resources = {}

# List of users as a tuple: (<<name>>, <<pswd>>, <<repeat count>>)
adminattrs = {
    "dsAttrTypeStandard:RealName": "Super User",
    "dsAttrTypeStandard:FirstName": "Super",
    "dsAttrTypeStandard:LastName": "User",
    "dsAttrTypeStandard:EMailAddress": "testadmin@example.com",
}

apprenticeattrs = {
    "dsAttrTypeStandard:RealName": "Apprentice Super User",
    "dsAttrTypeStandard:FirstName": "Apprentice",
    "dsAttrTypeStandard:LastName": "Super User",
    "dsAttrTypeStandard:EMailAddress": "apprentice@example.com",
}

userattrs = {
    "dsAttrTypeStandard:RealName": "User %02d",
    "dsAttrTypeStandard:FirstName": "User",
    "dsAttrTypeStandard:LastName": "%02d",
    "dsAttrTypeStandard:EMailAddress": "user%02d@example.com",
}

publicattrs = {
    "dsAttrTypeStandard:RealName": "Public %02d",
    "dsAttrTypeStandard:FirstName": "Public",
    "dsAttrTypeStandard:LastName": "%02d",
    "dsAttrTypeStandard:EMailAddress": "public%02d@example.com",
    "dsAttrTypeStandard:Street": "%d Public Row",
    "dsAttrTypeStandard:City": "Exampleville",
    "dsAttrTypeStandard:State": "Testshire",
    "dsAttrTypeStandard:PostalCode": "RFC 4791",
    "dsAttrTypeStandard:Country": "AAA",
}

i18nattrs = {
    "dsAttrTypeStandard:RealName": u"まだ",
    "dsAttrTypeStandard:FirstName": u"ま",
    "dsAttrTypeStandard:LastName": u"だ",
    "dsAttrTypeStandard:EMailAddress": "i18nuser@example.com",
}

locationcreatecmd = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
        <key>command</key>
        <string>createLocation</string>
        <key>AutoScheduleMode</key>
        <string>acceptIfFreeDeclineIfBusy</string>
        <key>GeneratedUID</key>
        <string>%(guid)s</string>
        <key>RealName</key>
        <string>%(realname)s</string>
        <key>RecordName</key>
        <array>
                <string>%(recordname)s</string>
        </array>
</dict>
</plist>
"""

locationremovecmd = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
        <key>command</key>
        <string>deleteLocation</string>
        <key>GeneratedUID</key>
        <string>%(guid)s</string>
</dict>
</plist>
"""

locationlistcmd = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
        <key>command</key>
        <string>getLocationList</string>
</dict>
</plist>
"""

resourcecreatecmd = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
        <key>command</key>
        <string>createResource</string>
        <key>AutoScheduleMode</key>
        <string>acceptIfFreeDeclineIfBusy</string>
        <key>GeneratedUID</key>
        <string>%(guid)s</string>
        <key>RealName</key>
        <string>%(realname)s</string>
        <key>Type</key>
        <string>Printer</string>
        <key>RecordName</key>
        <array>
                <string>%(recordname)s</string>
        </array>
        <key>Comment</key>
        <string>Test Comment</string>
</dict>
</plist>
"""

resourceremovecmd = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
        <key>command</key>
        <string>deleteResource</string>
        <key>GeneratedUID</key>
        <string>%(guid)s</string>
</dict>
</plist>
"""

resourcelistcmd = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
        <key>command</key>
        <string>getResourceList</string>
</dict>
</plist>
"""

locationattrs = {
    "dsAttrTypeStandard:RealName": "Location %02d",
}

delegatedroomattrs = {
    "dsAttrTypeStandard:RealName": "Delegated Conference Room",
}

resourceattrs = {
    "dsAttrTypeStandard:RealName": "Resource %02d",
}

groupattrs = {
    "dsAttrTypeStandard:RealName": "Group %02d",
    "dsAttrTypeStandard:EMailAddress": "group%02d@example.com",
}

records = (
    (kDSStdRecordTypeUsers, "testadmin", "testadmin", adminattrs, 1),
    (kDSStdRecordTypeUsers, "apprentice", "apprentice", apprenticeattrs, 1),
    (kDSStdRecordTypeUsers, "i18nuser", "i18nuser", i18nattrs, 1),
    (kDSStdRecordTypeUsers, "user%02d", "user%02d", userattrs, None),
    (kDSStdRecordTypeUsers, "public%02d", "public%02d", publicattrs, number_of_publics),
    (kDSStdRecordTypePlaces, "location%02d", "location%02d", locationattrs, number_of_locations),
    (kDSStdRecordTypePlaces, "delegatedroom", "delegatedroom", delegatedroomattrs, 1),
    (kDSStdRecordTypeResources, "resource%02d", "resource%02d", resourceattrs, number_of_resources),
    (kDSStdRecordTypeGroups, "group%02d", "group%02d", groupattrs, number_of_groups),
)

def usage():
    print """Usage: odsetup [options] create|create-users|remove
Options:
    -h        Print this help and exit
    -n node   OpenDirectory node to target
    -u uid    OpenDirectory Admin user id
    -p pswd   OpenDirectory Admin user password
    -c users  number of user accounts to create (default: 10)
    -x        disable OD node checks
    -v        verbose logging
    -V        very verbose logging
"""



def cmd(args, input=None, raiseOnFail=True):

    if veryverbose:
        print "-----"
    if verbose:
        print args.replace(diradmin_pswd, "xxxx")
    if veryverbose and input:
        print input
        print "*****"
    if input:
        p = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
        result = p.communicate(input)
    else:
        p = Popen(args, stdout=PIPE, stderr=PIPE, shell=True)
        result = p.communicate()

    if veryverbose:
        print "Output: %s" % (result[0],)
        print "Code: %s" % (p.returncode,)
    if raiseOnFail and p.returncode:
        raise RuntimeError(result[1])
    return result[0], p.returncode



def checkDataSource(node):
    """
    Verify that the specified node is the only node this host is bound to.

    @param node: the node to verify
    @type node: L{str}
    """

    result = cmd("dscl localhost -list /LDAPv3")
    result = ["/LDAPv3/{}".format(subnode) for subnode in result[0].splitlines()]
    if len(result) > 1 or result[0] != node:
        print "Error: Host is bound to other directory nodes: {}".format(result)
        print "CalDAVTester will likely fail with other nodes present."
        print "Please remove all nodes except the one being used for odsetup."
        sys.exit(1)



class ODError(Exception):
    pass



class ODFamework(object):

    def __init__(self, nodeName, user, pswd):
        self.session = ODSession.defaultSession()
        self.node, error = ODNode.nodeWithSession_name_error_(self.session, nodeName, None)
        if error:
            print(error)
            raise ODError(error)

        _ignore_result, error = self.node.setCredentialsWithRecordType_recordName_password_error_(
            kDSStdRecordTypeUsers,
            user,
            pswd,
            None
        )
        if error:
            print("Unable to authenticate with directory %s: %s" % (nodeName, error))
            raise ODError(error)

        print("Successfully authenticated with directory %s" % (nodeName,))


    def lookupRecordName(self, recordType, name):
        query, error = ODQuery.queryWithNode_forRecordTypes_attribute_matchType_queryValues_returnAttributes_maximumResults_error_(
            self.node,
            recordType,
            kDSNAttrRecordName,
            eDSExact,
            name,
            [kDS1AttrGeneratedUID],
            0,
            None)
        if error:
            raise ODError(error)
        records, error = query.resultsAllowingPartial_error_(False, None)
        if error:
            raise ODError(error)

        if len(records) < 1:
            return None
        if len(records) > 1:
            raise ODError("Multiple records for '%s' were found" % (name,))

        return records[0]


    def createRecord(self, recordType, recordName, password, attrs):
        record, error = self.node.createRecordWithRecordType_name_attributes_error_(
            recordType,
            recordName,
            attrs,
            None)
        if error:
            print(error)
            raise ODError(error)
        if recordType == kDSStdRecordTypeUsers:
            _ignore_result, error = record.changePassword_toPassword_error_(None, password, None)
            if error:
                print(error)
                raise ODError(error)
        return record


    def recordDetails(self, record):
        details, error = record.recordDetailsForAttributes_error_(None, None)
        if error:
            print(error)
            raise ODError(error)
        return details



def readConfig():
    """
    Read useful information from calendarserver_config
    """

    args = [
        configutility,
        "ServerHostName",
        "ConfigRoot",
        "DocumentRoot",
        "HTTPPort",
        "SSLPort",
        "Authentication.Basic.Enabled",
        "Authentication.Digest.Enabled",
    ]
    currentConfig = {}
    output, _ignore_code = cmd(" ".join(args), input=None)
    for line in output.split("\n"):
        if line:
            key, value = line.split("=")
            currentConfig[key] = value
    try:
        basic_ok = currentConfig["Authentication.Basic.Enabled"]
    except KeyError:
        basic_ok = False
    try:
        digest_ok = currentConfig["Authentication.Digest.Enabled"]
    except KeyError:
        digest_ok = False
    if digest_ok:
        authtype = "digest"
    elif basic_ok:
        authtype = "basic"

    return (
        currentConfig["ServerHostName"],
        int(currentConfig["HTTPPort"]),
        int(currentConfig["SSLPort"]),
        authtype,
        currentConfig["DocumentRoot"],
        currentConfig["ConfigRoot"],
    )



def patchConfig(confroot, admin):
    """
    Patch the caldavd-user.plist file to make sure:
       * the proper admin principal is configured
       * iMIP is disabled
       * SACLs are disabled
       * EnableAnonymousReadRoot is enabled
       * WorkQueue timings are appropriate for testing (low delay)

    @param admin: admin principal-URL value
    @type admin: str
    """
    plist = {}
    plist["AdminPrincipals"] = [admin]

    # For testing do not send iMIP messages!
    plist["Scheduling"] = {
        "iMIP" : {
            "Enabled" : False,
        },
    }

    # No SACLs
    plist["EnableSACLs"] = False

    # Needed for CDT
    plist["EnableAnonymousReadRoot"] = True
    plist["EnableControlAPI"] = True
    if "Options" not in plist["Scheduling"]:
        plist["Scheduling"]["Options"] = dict()

    # Lower WorkQueue timings to reduce processing delay
    plist["Scheduling"]["Options"]["WorkQueues"] = {
        "Enabled" : True,
        "RequestDelaySeconds" : 0.1,
        "ReplyDelaySeconds" : 1,
        "AutoReplyDelaySeconds" : 0.1,
        "AttendeeRefreshBatchDelaySeconds" : 0.1,
        "AttendeeRefreshBatchIntervalSeconds" : 0.1,
    }
    plist["WorkQueue"] = {
        "failureRescheduleInterval": 1,
        "lockRescheduleInterval": 1,
    }

    writePlist(plist, confroot + "/caldavd-user.plist")



def buildServerinfo(serverinfo_default, hostname, nonsslport, sslport, authtype, docroot):

    # Read in the serverinfo-template.xml file
    fd = open(serverinfo_template, "r")
    try:
        data = fd.read()
    finally:
        fd.close()

    subs_template = """
        <substitution>
            <key>%s</key>
            <value>%s</value>
        </substitution>
"""

    subs = [
        ("$useradminguid:", guids["testadmin"]),
        ("$userapprenticeguid:", guids["apprentice"]),
        ("$i18nguid:", guids["i18nuser"]),
    ]

    for i in range(1, number_of_users + 1):
        subs.append(("$userguid%d:" % (i,), guids["user%02d" % (i,)]))
    for i in range(1, number_of_publics + 1):
        subs.append(("$publicuserguid%d:" % (i,), guids["public%02d" % (i,)]))
    for i in range(1, number_of_resources + 1):
        subs.append(("$resourceguid%d:" % (i,), guids["resource%02d" % (i,)]))
    for i in range(1, number_of_locations + 1):
        subs.append(("$locationguid%d:" % (i,), guids["location%02d" % (i,)]))
    for i in range(1, number_of_groups + 1):
        subs.append(("$groupguid%d:" % (i,), guids["group%02d" % (i,)]))

    subs_str = ""
    for x, y in subs:
        subs_str += subs_template % (x, y,)

    data = data.format(
        hostname=hostname,
        nonsslport=str(nonsslport),
        sslport=str(sslport),
        authtype=authtype,
        overrides=subs_str,
        DAV="{DAV:}",
    )

    fd = open(serverinfo_default, "w")
    try:
        fd.write(data)
    finally:
        fd.close()



def loadLists(path, records):
    if path == kDSStdRecordTypePlaces:
        result = cmd(cmdutility, locationlistcmd)
    elif path == kDSStdRecordTypeResources:
        result = cmd(cmdutility, resourcelistcmd)
    else:
        raise ValueError()

    try:
        plist = readPlistFromString(result[0])
    except xml.parsers.expat.ExpatError, e:
        print "Error (%s) parsing (%s)" % (e, result[0])
        raise

    for record in plist["result"]:
        records[record["RecordName"][0]] = record["GeneratedUID"]



def doToAccounts(odf, protocol, f, users_only=False):

    for record in records:
        if protocol == "carddav" and record[0] in (kDSStdRecordTypePlaces, kDSStdRecordTypeResources):
            continue
        if record[4] is None:
            count = number_of_users
        elif users_only:
            continue
        else:
            count = record[4]
        if count > 1:
            for ctr in range(1, count + 1):
                attrs = {}
                for key, value in record[3].iteritems():
                    if value.find("%02d") != -1:
                        value = value % (ctr,)
                    attrs[key] = value
                ruser = (record[1] % (ctr,), record[2] % (ctr,), attrs, 1)
                f(odf, record[0], ruser)
        else:
            f(odf, record[0], record[1:])



def doGroupMemberships(odf):

    memberships = (
        ("group01", ("user01",), (),),
        ("group02", ("user06", "user07",), (),),
        ("group03", ("user08", "user09",), (),),
        ("group04", ("user10",), ("group02", "group03",),),
        ("group05", ("user20",), ("group06",),),
        ("group06", ("user21",), (),),
        ("group07", ("user22", "user23", "user24",), (),),
    )

    for groupname, users, nestedgroups in memberships:
        if verbose:
            print "Group membership: {}".format(groupname)

        # Get group record
        group = odf.lookupRecordName(kDSStdRecordTypeGroups, groupname)
        if group is not None:
            for user in users:
                member = odf.lookupRecordName(kDSStdRecordTypeUsers, user)
                if member is not None:
                    _ignore_result, error = group.addMemberRecord_error_(member, None)
                    if error:
                        raise ODError(error)
            for nested in nestedgroups:
                member = odf.lookupRecordName(kDSStdRecordTypeGroups, nested)
                if member is not None:
                    _ignore_result, error = group.addMemberRecord_error_(member, None)
                    if error:
                        raise ODError(error)



def createUser(odf, path, user):

    if verbose:
        print "Create user: {}/{}".format(path, user[0])

    if path in (kDSStdRecordTypeUsers, kDSStdRecordTypeGroups,):
        createUserViaDS(odf, path, user)
    elif protocol == "caldav":
        createUserViaGateway(path, user)



def createUserViaDS(odf, path, user):
    # Do dscl command line operations to create a calendar user

    # Only create if it does not exist
    record = odf.lookupRecordName(path, user[0])

    if record is None:
        # Create the user
        if kDS1AttrGeneratedUID in user[2]:
            user[2][kDS1AttrGeneratedUID] = str(uuid.uuid4()).upper()

        user = (user[0], user[1], dict([(k, [v],) for k, v in user[2].items()]),)
        record = odf.createRecord(path, user[0], user[1], user[2])
    else:
        if verbose:
            print "%s/%s already exists" % (path, user[0],)

    # Now read the guid for this record
    if user[0] in guids:
        record = odf.lookupRecordName(path, user[0])
        details = odf.recordDetails(record)
        guids[user[0]] = details[kDS1AttrGeneratedUID][0]



def createUserViaGateway(path, user):

    # Check for existing
    if path == kDSStdRecordTypePlaces:
        if user[0] in locations:
            guids[user[0]] = locations[user[0]]
            return
    elif path == kDSStdRecordTypeResources:
        if user[0] in resources:
            guids[user[0]] = resources[user[0]]
            return

    guid = str(uuid.uuid4()).upper()
    if user[0] in guids:
        guids[user[0]] = guid
    if path == kDSStdRecordTypePlaces:
        cmd(
            cmdutility,
            locationcreatecmd % {
                "guid": guid,
                "realname": user[2]["dsAttrTypeStandard:RealName"],
                "recordname": user[0]
            }
        )
    elif path == kDSStdRecordTypeResources:
        cmd(
            cmdutility,
            resourcecreatecmd % {
                "guid": guid,
                "realname": user[2]["dsAttrTypeStandard:RealName"],
                "recordname": user[0]
            }
        )
    else:
        raise ValueError()



def removeUser(odf, path, user):

    if verbose:
        print "Remove user: {}/{}".format(path, user[0])

    if path in (kDSStdRecordTypeUsers, kDSStdRecordTypeGroups,):
        removeUserViaDS(odf, path, user)
    else:
        removeUserViaGateway(path, user)



def removeUserViaDS(odf, path, user):
    # Do dscl command line operations to remove a calendar user

    record = odf.lookupRecordName(path, user[0])
    if record is not None:
        _ignore_result, error = record.deleteRecordAndReturnError_(None)
        if error:
            raise ODError(error)



def removeUserViaGateway(path, user):

    if path == kDSStdRecordTypePlaces:
        if user[0] not in locations:
            return
        guid = locations[user[0]]
        cmd(
            cmdutility,
            locationremovecmd % {"guid": guid, }
        )
    elif path == kDSStdRecordTypeResources:
        if user[0] not in resources:
            return
        guid = resources[user[0]]
        cmd(
            cmdutility,
            resourceremovecmd % {"guid": guid, }
        )
    else:
        raise ValueError()



def manageRecords(odf, path, user):
    """
    Set proxies and auto-schedule for locations and resources
    """

    # Do caldav_utility setup
    if path in (kDSStdRecordTypePlaces, kDSStdRecordTypeResources,):
        if path in (kDSStdRecordTypePlaces,):
            if user[0] == "delegatedroom":
                cmd("%s --add-write-proxy groups:group05 --add-read-proxy groups:group07 --set-auto-schedule-mode=none locations:%s" % (
                    utility,
                    user[0],
                ))
            else:
                cmd("%s --add-write-proxy users:user01 --set-auto-schedule-mode=automatic locations:%s" % (
                    utility,
                    user[0],
                ))
        else:
            # Default options for all resources
            cmd("%s --add-write-proxy users:user01 --add-read-proxy users:user03 --set-auto-schedule-mode=automatic resources:%s" % (
                utility,
                user[0],
            ))

            # Some resources have unique auto-schedule mode set
            automodes = {
                "resource05" : "none",
                "resource06" : "accept-always",
                "resource07" : "decline-always",
                "resource08" : "accept-if-free",
                "resource09" : "decline-if-busy",
                "resource10" : "automatic",
                "resource11" : "decline-always",
            }

            if user[0] in automodes:
                cmd("%s --set-auto-schedule-mode=%s resources:%s" % (
                    utility,
                    automodes[user[0]],
                    user[0],
                ))

            # Some resources have unique auto-accept-groups assigned
            autoAcceptGroups = {
                "resource11" : "group01",
            }
            if user[0] in autoAcceptGroups:
                cmd("%s --set-auto-accept-group=groups:%s resources:%s" % (
                    utility,
                    autoAcceptGroups[user[0]],
                    user[0],
                ))


if __name__ == "__main__":

    protocol = "caldav"
    serverinfo_default = details[protocol]["serverinfo"]
    node_check = True
    try:
        options, args = getopt.getopt(sys.argv[1:], "hn:p:u:f:c:vVx")

        for option, value in options:
            if option == "-h":
                usage()
                sys.exit(0)
            elif option == "-n":
                directory_node = value
            elif option == "-u":
                diradmin_user = value
            elif option == "-p":
                diradmin_pswd = value
            elif option == "-c":
                number_of_users = int(value)
            elif option == "-v":
                verbose = True
            elif option == "-V":
                verbose = True
                veryverbose = True
            elif option == "-x":
                node_check = False
            else:
                print "Unrecognized option: %s" % (option,)
                usage()
                raise ValueError

        if not diradmin_pswd:
            diradmin_pswd = getpass("Directory Admin Password: ")

        # Process arguments
        if len(args) == 0:
            print "No arguments given. One of 'create' or 'remove' must be present."
            usage()
            raise ValueError
        elif len(args) > 1:
            print "Too many arguments given. Only one of 'create' or 'remove' must be present."
            usage()
            raise ValueError
        elif args[0] not in ("create", "create-users", "remove"):
            print "Wrong arguments given: %s" % (args[0],)
            usage()
            raise ValueError

        odf = ODFamework(directory_node, diradmin_user, diradmin_pswd)

        if node_check:
            checkDataSource(directory_node)

        if args[0] == "create":
            # Read the caldavd.plist file and extract some information we will need.
            hostname, port, sslport, authtype, docroot, confroot = readConfig()

            # Now generate the OD accounts (caching guids as we go).
            if protocol == "caldav":
                loadLists(kDSStdRecordTypePlaces, locations)
                loadLists(kDSStdRecordTypeResources, resources)

            doToAccounts(odf, protocol, createUser)
            doGroupMemberships(odf)
            doToAccounts(odf, protocol, manageRecords)

            # Patch the caldavd.plist file with the testadmin user's guid-based principal-URL
            patchConfig(confroot, "/principals/__uids__/%s/" % (guids["testadmin"],))

            # Create an appropriate serverinfo.xml file from the template
            buildServerinfo(serverinfo_default, hostname, port, sslport, authtype, docroot)


        elif args[0] == "create-users":
            # Read the caldavd.plist file and extract some information we will need.
            hostname, port, sslport, authtype, docroot, confroot = readConfig()

            # Now generate the OD accounts (caching guids as we go).
            if protocol == "caldav":
                loadLists(kDSStdRecordTypePlaces, locations)
                loadLists(kDSStdRecordTypeResources, resources)

            doToAccounts(odf, protocol, createUser, users_only=True)

            # Create an appropriate serverinfo.xml file from the template
            buildServerinfo(serverinfo_default, hostname, port, sslport, authtype, docroot)

        elif args[0] == "remove":
            if protocol == "caldav":
                loadLists(kDSStdRecordTypePlaces, locations)
                loadLists(kDSStdRecordTypeResources, resources)
            doToAccounts(odf, protocol, removeUser)

    except Exception, e:
        traceback.print_exc()
        sys.exit(1)



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Verifier that checks the response body for an exact match to data in a file.
"""

from StringIO import StringIO
from difflib import unified_diff
from xml.etree.cElementTree import ElementTree, tostring
import os

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable
        # Get arguments
        files = args.get("filepath", [])
        if manager.data_dir:
            files = map(lambda x: os.path.join(manager.data_dir, x), files)

        # status code must be 200, 207
        if response.status not in (200, 207):
            return False, "        HTTP Status Code Wrong: %d" % (response.status,)

        # look for response data
        if not respdata:
            return False, "        No response body"

        # look for one file
        if len(files) != 1:
            return False, "        No file to compare response to"

        # read in all data from specified file
        fd = open(files[0], "r")
        try:
            try:
                data = fd.read()
            finally:
                fd.close()
        except:
            data = None

        if data is None:
            return False, "        Could not read data file"

        data = manager.server_info.subs(data)

        result = True
        if data != respdata:
            data = data.replace("\n", "\r\n")
            if data != respdata:
                # If we have an iCalendar file, then unwrap data and do compare
                if files[0].endswith(".ics"):
                    data = data.replace("\r\n ", "")
                    respdata = respdata.replace("\r\n ", "")
                    if data != respdata:
                        result = False
                elif files[0].endswith(".xml"):
                    try:
                        respdata = tostring(ElementTree(file=StringIO(respdata)).getroot())
                    except Exception:
                        return False, "        Could not parse XML response: %s" % (respdata,)
                    try:
                        data = tostring(ElementTree(file=StringIO(data)).getroot())
                    except Exception:
                        return False, "        Could not parse XML data: %s" % (data,)
                    if data != respdata:
                        result = False
                else:
                    result = False

        if result:
            return True, ""
        else:
            error_diff = "\n".join([line for line in unified_diff(data.split("\n"), respdata.split("\n"))])
            return False, "        Response data does not exactly match file data %s" % (error_diff,)



##
# Copyright (c) 2007-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
XML processing utilities.
"""

import src.xmlDefs

def readStringElementList(node, ename):

    results = []
    for child in node.getchildren():
        if child.tag == ename:
            results.append(child.text.decode("utf-8"))
    return results



def getYesNoAttributeValue(node, attr):
    return node.get(attr, src.xmlDefs.ATTR_VALUE_NO) == src.xmlDefs.ATTR_VALUE_YES



def getDefaultAttributeValue(node, attr, default):
    result = node.getAttribute(attr)
    if result:
        return result
    else:
        return default



def readOneStringElement(node, ename):

    for child in node.getchildren():
        if child.tag == ename:
            return child.text.decode("utf-8")
    return ""



def nodeForPath(root, path):
    if '[' in path:
        actual_path, tests = path.split('[', 1)
    else:
        actual_path = path
        tests = None

    # Handle absolute root element
    if actual_path[0] == '/':
        actual_path = actual_path[1:]
    if '/' in actual_path:
        root_path, child_path = actual_path.split('/', 1)
        if root.tag != root_path:
            return None
        nodes = root.findall(child_path)
    else:
        root_path = actual_path
        child_path = None
        nodes = (root,)

    if len(nodes) == 0:
        return None

    results = []

    if tests:
        tests = [item[:-1] for item in tests.split('[')]
        for test in tests:
            for node in nodes:
                if test[0] == '@':
                    if '=' in test:
                        attr, value = test[1:].split('=')
                        value = value[1:-1]
                    else:
                        attr = test[1:]
                        value = None
                    if attr in node.keys() and (value is None or node.get(attr) == value):
                        results.append(node)
                elif test[0] == '=':
                    if node.text == test[1:]:
                        results.append(node)
                elif test[0] == '!':
                    if node.text != test[1:]:
                        results.append(node)
                elif test[0] == '*':
                    if node.text is not None and node.text.find(test[1:]) != -1:
                        results.append(node)
                elif test[0] == '+':
                    if node.text is not None and node.text.startswith(test[1:]):
                        results.append(node)
                elif test[0] == '^':
                    if "=" in test:
                        element, value = test[1:].split("=", 1)
                    else:
                        element = test[1:]
                        value = None
                    for child in node.getchildren():
                        if child.tag == element and (value is None or child.text == value):
                            results.append(node)
    else:
        results = nodes

    return results



def xmlPathSplit(xpath):

    pos = xpath.find("}")
    if pos == -1:
        return xpath, ""
    pos = xpath[pos:].find("/") + pos
    if pos == -1:
        return xpath, ""
    else:
        return xpath[:pos], xpath[pos + 1:]



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Verifier that checks the response for a pre/post-condition <DAV:error> result.
"""

from xml.etree.cElementTree import ElementTree
from StringIO import StringIO

class Verifier(object):

    def verify(self, manager, uri, response, respdata, args): #@UnusedVariable
        # If no status verification requested, then assume all 2xx codes are OK
        teststatus = args.get("error", [])
        statusCode = args.get("status", ["403", "409", "507"])
        ignoreextras = args.get("ignoreextras", None)

        # status code could be anything, but typically 403, 409 or 507
        if str(response.status) not in statusCode:
            return False, "        HTTP Status Code Wrong: %d" % (response.status,)

        # look for pre-condition data
        if not respdata:
            return False, "        No pre/post condition response body"

        try:
            tree = ElementTree(file=StringIO(respdata))
        except Exception, ex:
            return False, "        Could not parse XML: %s" % (ex,)

        if tree.getroot().tag != "{DAV:}error":
            return False, "        Missing <DAV:error> element in response"

        # Make a set of expected pre/post condition elements
        expected = set(teststatus)
        got = set()
        for child in tree.getroot().getchildren():
            if child.tag != "{http://twistedmatrix.com/xml_namespace/dav/}error-description":
                got.add(child.tag)

        missing = expected.difference(got)
        extras = got.difference(expected)

        err_txt = ""
        if len(missing):
            err_txt += "        Items not returned in error: element %s" % str(missing)
        if len(extras) and not ignoreextras:
            if len(err_txt):
                err_txt += "\n"
            err_txt += "        Unexpected items returned in error element: %s" % str(extras)
        if len(missing) or len(extras) and not ignoreextras:
            return False, err_txt

        return True, ""



##
# Copyright (c) 2006-2015 Apple Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##

"""
Class that encapsulates a single CalDAV test.
"""

from src.request import request
from src.xmlUtils import getYesNoAttributeValue
import src.xmlDefs

class test(object):
    """
    A single test which can be comprised of multiple requests. The test can
    be run more than once, and timing information gathered and averaged across
    all runs.
    """

    def __init__(self, manager):
        self.manager = manager
        self.name = ""
        self.details = False
        self.count = 1
        self.stats = False
        self.ignore = False
        self.only = False
        self.require_features = set()
        self.exclude_features = set()
        self.description = ""
        self.requests = []


    def missingFeatures(self):
        return self.require_features - self.manager.server_info.features


    def excludedFeatures(self):
        return self.exclude_features & self.manager.server_info.features


    def parseXML(self, node):
        self.name = node.get(src.xmlDefs.ATTR_NAME, "")
        self.details = getYesNoAttributeValue(node, src.xmlDefs.ATTR_DETAILS)
        self.count = int(node.get(src.xmlDefs.ATTR_COUNT, 1))
        self.stats = getYesNoAttributeValue(node, src.xmlDefs.ATTR_STATS)
        self.ignore = getYesNoAttributeValue(node, src.xmlDefs.ATTR_IGNORE)
        self.only = getYesNoAttributeValue(node, src.xmlDefs.ATTR_ONLY)

        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_REQUIRE_FEATURE:
                self.parseFeatures(child, require=True)
            elif child.tag == src.xmlDefs.ELEMENT_EXCLUDE_FEATURE:
                self.parseFeatures(child, require=False)
            elif child.tag == src.xmlDefs.ELEMENT_DESCRIPTION:
                self.description = child.text

        # get request
        self.requests = request.parseList(self.manager, node)


    def parseFeatures(self, node, require=True):
        for child in node.getchildren():
            if child.tag == src.xmlDefs.ELEMENT_FEATURE:
                (self.require_features if require else self.exclude_features).add(child.text.encode("utf-8"))


    def dump(self):
        print "\nTEST: %s" % self.name
        print "    description: %s" % self.description
        for req in self.requests:
            req.dump()
